import { RequestOperacaoDataset } from "../nRequests";
import CustomStore from 'devextreme/data/custom_store';
import { formatWithDataType } from "../nGrid";
import { handleResponseBox } from "./actions/messageBox/handleReponseBox";

/**
 * Implement CRUD operations on plain array of items. 
 * Operations are performed first on array and then 
 * propagated to the server.
 */
export class CustomStoreArray extends CustomStore {
	constructor(args) {
		super({
			...args,
			byKey: (key) => this._getRowByKey(key),
			load: () => this._dataSourceArray,
			totalCount: () => this._dataSourceArray.length,
			update: (key, values) => this._updateData(key, values),
			insert: (data) => this._insertNewItem(data),
		});

		const { dataSourceArray, Class, Params, Name, requestDataType, datasetNames, columnOptions } = args.customStoreArrayOptions
		
		this._dataSourceArray = dataSourceArray;
		this._Class = Class;
		this._Name = Name;
		this._params = Params;
		this._requestDataType = requestDataType;
		this._datasetNames = datasetNames;
		this._columnOptions = this._createColumnOptions(columnOptions);
		this._keyStore = this._createKeyStore(this._dataSourceArray);
		this._insertOperations = new Map();
		this._promiseSender = null;
		this._dataToUpdate = [];
	}

	_createColumnOptions(options) {
		let map = new Map();

		options.forEach((option) => {
			map.set(option.dataField, option);
		});

		return map;
	}

	setDataSourceArray(data) {
		this._dataSourceArray = data;
	}

	/**
	 * 
	 * @param {[]} data Array with rowData
	 */
	setDataToUpdate(data) {
		this._dataToUpdate = data.map((item) => {
			return {
				key: this.keyOf(item),
				values: item
			}
		})
	}

	_getRowByKey(key) {
		let hash = this._getHashObj(key);

		return this._dataSourceArray[this._keyStore.get(hash)];
	}

	_createKeyStore(items) {
		let keyStore = new Map();

		items.forEach((item, index) => {
			let hash = this._getHashObj(item);

			if (!keyStore.has(hash)) keyStore.set(hash, index)
			else {
				let value = keyStore.get(hash);

				keyStore.set(hash, value.length ? (value.push(index), value) : [value, index])
			}
		})

		return keyStore;
	}

	/**
	 * 
	 * @param {string|number|object} item 
	 * 
	 * Get an string representation of an object. If the key has only one column, convert key to match string representaion
	 */
	_getHashObj(item) {
		let itemObj = item instanceof Object ? item : { [this._key]: item };

		return JSON.stringify(this.keyOf(itemObj));
	}

	/**
	 * 
	 * @param {Object} data { key, value }
	 * 
	 * Return a single promise which will execute a server call in the next execution of the event loop
	 * after all the modified data has been collected 
	 */
	_mountBach(data) {
		this._dataToUpdate.push(data);
		
		if (!this._promiseSender) {
			this._promiseSender = new Promise((resolve, reject) => {
				setTimeout(() => this._request({ resolve, reject }), 2000)
			});
		}

		return this._promiseSender;
	}

	_request(promHandler) {
		let data = this._getDataForRequest();

		this._dataToUpdate = [];
		this._promiseSender = null;

		this._makeRequest(data, promHandler);
	}

	_getDataForRequest() {
		const getModifiedDatasource = (data, dataToUpdate) => {
			let newData = data.map((item) => ({...item}));
			
			dataToUpdate.forEach(({ keys, values }) => {
				let hash = this._getHashObj(keys);

				if (this._keyStore.has(hash)) { // If exists, handle collision
					let value = this._keyStore.get(hash);
					let indexes = value.length ? value : [value];

					indexes.forEach((index) => { 
						newData[index] = {...newData[index], ...values}; 
					});
				}
				else if (this._insertOperations.has(hash)) { // new item
					newData.push(values);
				}
			});
			
			return newData;
		};
		const fixValues = (data) => {
			return data.map(row => {
				const fixedRow = { ...row };

				Object
					.entries(fixedRow)
					.forEach(([columnName, columnValue]) => {
						let { dataType } = this._columnOptions.get(columnName);

						formatWithDataType(dataType, fixedRow, columnName, columnValue);
					});

				return fixedRow;
			});
		};
		const getDataToUpdate = () => {
			switch(this._requestDataType) {
				case 'dsUpdated': return fixValues(this._dataToUpdate.map(({ key, values }) => (values)));
				case 'dsAll': return fixValues(getModifiedDatasource(this._dataSourceArray, this._dataToUpdate));
				case 'dsSelected': console.error('"dsSelected" is not an valid option for DataSend');
				default: return [];
			}
		};

		return [{
			NomeDataset: this._datasetNames[0] || '',
			Registros: getDataToUpdate(),
		}]
	}

	_makeRequest(data, promHandler) {
		console.log('Request data:', data);

		let request = new RequestOperacaoDataset({
			_ADatamodule: this._Class,
			_AAction: this._Name,
			_ADataset: data,
			_AProcessoParams: this._params
		});

		request.executar(this._handleResponse.bind(this), promHandler);
	}

	_handleResponse(response, { resolve, reject }) {
		const { data: { Action, ClassName, Dataset, Erro, dialogs, MsgTreeLista, MsgLista } } = response;
		const { DataReturn: responseType, Registros, NomeDataset } = Dataset[0];

		console.log('response', response);

		if (Erro) {
			reject(Erro)
		}
		else {
			switch (responseType) {
				case 'drAll':
					this._handleResponseAll(Registros);
					break;
				case 'drUpdated':
					this._handleResponseUpdated(Registros);
					break;
			}
			resolve()
		}
		handleResponseBox(NomeDataset, { MsgTreeLista, MsgLista });
	}

	_handleResponseAll(data) {
		this._dataSourceArray = data;
		this._keyStore = this._createKeyStore(this._dataSourceArray);

		this._insertOperations.clear();
	}

	_handleResponseUpdated(data) {
		data.forEach((item) => {
			const hash = this._getHashObj(item)

			if (this._keyStore.has(hash)) {
				this._dataSourceArray[this._keyStore.get(hash)] = item;
			}
			else if (this._insertOperations.has(hash)) {
				this._insertOperations.delete(hash);

				this._keyStore.set(hash, this._dataSourceArray.length);
				
				this._dataSourceArray.push(item);
			}
		});
	}

	_updateData(keys, values) {
		let newValues = { ...(this._getRowByKey(keys) || {}), ...values };

		return this._mountBach({ keys, values: newValues });
	}

	async _insertNewItem(data) {
		const newIndex = this._dataSourceArray.length + this._insertOperations.size + 1;

		const newData = {
			...this._getCleanItem(),
			...data,
			INDICE: `${newIndex}`,
		};

		const hash = this._getHashObj(newData);

		this._insertOperations.set(hash, true);

		try {
			return this._mountBach({
				keys: this.keyOf(newData),
				values: newData,
			});
		} 
		catch (e) {
			this._insertOperations.clear();

			throw e; //raise error to dataSource
		}
	}

	_getCleanItem() {
		const cleanItem = {};
		const columnNames = this._columnOptions.keys();

		for (const name of columnNames) {
			if (name) cleanItem[name] = '';
		}

		return cleanItem;
	}
}
