import { Subject } from "rxjs";
import { makeRequestLookupValue } from '../nLookUp';
import dxTextBox from 'devextreme/ui/text_box';
import dxDateBox from 'devextreme/ui/date_box';
import { nExpressionExecutor } from "../nExpressionExecutor";
import { GridExpressions } from "./GridExpressions";
import { DateConfigurator } from "./DateConfigurator";
import { GetModalDataTableSelectedColumnSingleValue, GetModalDataTableSelectedColumnValues } from "../nWebUtil";


export const createGridLookup = () => {
	return (cellElement, cellInfo) => {
		const { column: { dataType, mask }, value } = cellInfo;
		const expressionExecutor = new nExpressionExecutor();
		const commonEditorOptions = {
			mask,
			value,
			onValueChanged:  editorOnChange(cellInfo, expressionExecutor),
			buttons: [{
				name: 'lookup',
				location: 'after',
				options: {
					template: templateSearchButton(),
					onClick: onClickSearchButton(cellInfo),
					type: 'normal',
					icon: 'la la-search',
				}
			}]
		};

		switch(dataType) {
			case 'date': return getDateBoxElement(commonEditorOptions);
			default: return getTextBoxElement(commonEditorOptions);
		}
	}
};

const getDateBoxElement = ({ buttons, onValueChanged, value, mask }) => {
	let dateBox = new dxDateBox($('<div>'), {
		value,
		buttons, // buttons: ['dropDown'], // datePicker button
		...DateConfigurator.getEditorOptions(mask, { onValueChanged }),
	});

	return dateBox.element();
}

const getTextBoxElement = (options) => {
	let textBox = new dxTextBox($('<div>'), {	...options });

	return textBox.element();
}


const editorOnChange = (cellInfo, expressionExecutor) => {
	const { column, data, setValue, component, rowIndex } = cellInfo;
	const { dataField, Lookup } = column;
	const { isNtcLinkAnnotation } = Lookup;

	const setDescriptionTrigger = (text) => {
		const resultFields = Lookup.LinkLookupResultField
			.split('+')
			.reduce((fields, fieldName) => {
				return [
					...fields,
					{ 
						Name: fieldName,
						Value: '',
					},
				];
			}, []);

		setTimeout(() => {
			setDescription({
				data: {
					[text ? Lookup.FieldLookup : 'Fields']: text ? text : resultFields,
				}
			}, cellInfo);
		})
	};

	const onChangeExpression = GridExpressions.onChange({
		data,
		rowIndex,
		expressionExecutor,
		editorField: dataField,
		gridComponent: component,
	});

	return async (e) => {
		const { event, element, component } = e;
		let newValue = event.target.value;

		onChangeExpression(e);

		if (isNtcLinkAnnotation) return;

		if (newValue) {
			try {
				let response = await FindGridLookupValue({ data, newValue, Lookup, _AName: dataField, _AAlias: '' });
				setDescription(response, cellInfo);								
			}
			catch {
				setDescriptionTrigger('Falha ao buscar lookup...');
			}
		}
		else {
			setDescriptionTrigger('');
		}
	}
};

const setDescription = ({ data }, cellInfo) => {
	let { column, rowIndex, component } = cellInfo;	
	
    const fixDateValue = (fieldName, value) => {
		const filterDateValue = (value, date) => (value.includes('Registro não encontrado') || value === '' ? '' : date);
        const { dataType } = component.columnOption(fieldName);

        switch (dataType) {
            case 'date': return filterDateValue(value, moment(value, 'DD/MM/YYYY').format('MM/DD/YYYY'));
            case 'datetime': return filterDateValue(value, moment(value, 'DD/MM/YYYY h:m:s').format('MM/DD/YYYY HH:m:s'));
            default: return value;                
        }
    }	
	
	if (data.Fields) {
		data.Fields.forEach((field) => {
			const value = fixDateValue(field.Name, field.Value);

			component.cellValue(rowIndex, field.Name, value);
		});	
	} else {
		const { FieldLookup } = column.Lookup;
		const value = fixDateValue(FieldLookup, data[FieldLookup]);
		component.cellValue(rowIndex, FieldLookup, value)
	}
};

function FindGridLookupValue({ _AName, _AAlias, data, newValue, Lookup }) {
	const { Cadastro, KeyFieldControls, FieldLookup, LinkKeyFields, LinkLookupDataSetNome, LinkLookupKeyFields, LinkLookupResultField } = Lookup;

	let _AFieldsValue = getGridKeyValues({
		newValue,
		data,
		LinkLookupKeyFields,
		KeyFieldControls,
		LinkKeyFields,
		Name: _AName,
	});

	return makeRequestLookupValue({
		_AAlias,
		_AFieldsValue,
		LinkLookupDataSetNome,
		LinkLookupResultField,
		_AControleResult: FieldLookup,
		_ADMClassName: Cadastro,
	})
}

function getGridKeyValues({ newValue, data, LinkLookupKeyFields, KeyFieldControls, LinkKeyFields, Name }) {
	let _linkLookupKeyFields = LinkLookupKeyFields.split(';');
	let _linkKeyFields = (KeyFieldControls || LinkKeyFields).split(';');

	return _linkLookupKeyFields.reduce((keyValues, FieldName, index) => {
		let isLast = (index == (_linkLookupKeyFields.length - 1));
		let searchField = isLast ? Name : _linkKeyFields[index];

		keyValues.push({
			FieldName,
			Value: isLast ? newValue : data[searchField]
		});

		return keyValues;
	}, []);
}

const templateSearchButton = () => (buttonData, contentElement) => {
	contentElement.parent().css({
		'border-radius': 0,
		'border-bottom': 0,
		'border-top': 0,
		'margin': 0,
		'height': '100%'
	});

	contentElement.css({
		'display': 'flex',
		'align-items': 'center',
		'padding': '2px 1rem'
	});

	return $(`<i class="${buttonData.icon}"></i>`);
};

const onClickSearchButton = (cellInfo) => {
	const { data, column: { Lookup } } = cellInfo;

	return () => {
		let lookupResponse = new Subject();

		lookupResponse.subscribe((selectedRows) => { handleSelection(selectedRows, cellInfo) });

		FormManager.QuemChamou = {
			Lookup,
			isFromGridLookup: true,
			observer: lookupResponse,
			data
		};

		$('#myModal').modal('show');
	}
};

function handleSelection(selectedRows, cellInfo) {
	const { column, data, component, rowIndex } = cellInfo;
	const { Lookup: { LinkLookupKeyFields, LinkLookupResultField, FieldLookup, isNtcLinkAnnotation, LinkKeyFields } } = column;
	const expressionExecutor = new nExpressionExecutor();

	const Name = column.dataField;
	const arraylinklookupkeyfields = LinkLookupKeyFields.split(';');
	const arraylinkKeyfields = LinkKeyFields.split(';')

	arraylinklookupkeyfields
		.forEach((linkLookupKeyField, index) => {
		const getColumnValues = (_AColumnName) => GetModalDataTableSelectedColumnValues({ _AColumnName, selectedRows });
		const getColumnSingleValues = (_AColumnName) => GetModalDataTableSelectedColumnSingleValue({ selectedRows, _AColumnName });

		const isLast = (index == (arraylinklookupkeyfields.length - 1));
		const value = isLast ? getColumnValues(linkLookupKeyField) : getColumnSingleValues(linkLookupKeyField);
		const itemLinkKeyField = isLast ? Name : arraylinkKeyfields[index];

		const onHandleExpression = GridExpressions.onHandleSelection({
			data,
			rowIndex,
			expressionExecutor,
			editorField: column.dataField,
			gridComponent: component,
			valueField: value
		});

		component.cellValue(rowIndex, itemLinkKeyField, isNtcLinkAnnotation ? `${data[Name]}${value}` : value);

		onHandleExpression();

		if (LinkLookupResultField) {
			if (LinkLookupResultField.includes("+")) {
				LinkLookupResultField
					.split('+')
					.forEach((fieldName) => {
						const description = getColumnSingleValues(fieldName);

						component.cellValue(rowIndex, fieldName, description);
					});
			} else {
				const description = (selectedRows.length > 1)
					? 'Seleção Multipla'
					: LinkLookupResultField.split(';')
						.map((columnName) => getColumnSingleValues(columnName))
						.join(' | ');

				component.cellValue(rowIndex, FieldLookup, description);
			}

			setTimeout(() => component.editCell(rowIndex, Name));
		};
	});
}