import axios, { AxiosRequestConfig } from "axios";
import { Subject } from "rxjs";
import { first, map } from "rxjs/operators";
import { v4 as uuidv4 } from "uuid";
import { verifyAxiosError, verifyAxiosRedirect } from "../../js/nRequests";
import { bufferDebounce } from "../components/widgets/util/buffer-debounce";
import { GetGridConfigConfig, GetGridConfigType } from "./services/get-grid-config";
import { DadosLookupDebounceConfig, RequestDadosLookupDebounceType } from "./services/request-dados-lookup";

interface DebounceResponse<T = any> {
    id: string;
    payload: T;
}

interface DebounceRequest<T = any>  {
    id: string; // randomId
    requestData: RequestData<T>;
}

interface RequestData<T = any> { 
    urlParams?: T; // == axios params
    bodyData?: T; // == axios data 
}

type RequestType = 
    | GetGridConfigType
    | RequestDadosLookupDebounceType;

export interface SetRequestDebounceConfig<T extends RequestType = RequestType> {
    type: T;
    config: DebounceRequestConfig;
}

type DebounceRequestConfig = {
    axiosConfig: AxiosRequestConfig;
    debounceTime?: number;
    useURLSearchParams?: boolean;
    arrayPropName: string;
}

class Api {
    private _debounceRequests: Map<RequestType, Subject<DebounceRequest>>;
    private _debounceResponse: Subject<DebounceResponse>;
    private _debounceRequestConfig: Map<RequestType, DebounceRequestConfig>;

    constructor() {
        this._debounceRequests = new Map<RequestType, Subject<DebounceRequest>>();
        this._debounceResponse = new Subject<DebounceResponse>();
        this._debounceRequestConfig = new Map<RequestType, DebounceRequestConfig>();

        this._initRequestConfig();
    }

    private _initRequestConfig() {
        this._setRequestConfigType(DadosLookupDebounceConfig());
        this._setRequestConfigType(GetGridConfigConfig());
    }

    private _setRequestConfigType({ type, config }: SetRequestDebounceConfig) {
        this._debounceRequestConfig.set(type, config);
    }    

    public requestWithDebounce<R = any, T = any>({ type, requestData } : { type: RequestType, requestData: RequestData<T> }) {
        const uuid = uuidv4();
        const requestConfig = this._debounceRequestConfig.get(type);

        if (!requestConfig) throw new Error(`Request not configured for type ${type}`);

        if (!this._debounceRequests.has(type)) {
            const debouncer = new Subject<DebounceRequest>();

            this._debounceRequests.set(type, debouncer);

            debouncer
                .pipe(bufferDebounce(requestConfig.debounceTime || 200))
                .subscribe((requests) => this._createRequestWithDebounce(requestConfig, requests));
        }

        this._debounceRequests
            .get(type)
            ?.next({
                id: uuid,
                requestData
            });

        return this._debounceResponse
            .pipe(
                first((response) => response.id == uuid),
                map((response) => response.payload)
            )
            .toPromise<R>()
            .then<R>((response: any) => {
                if (response?.apiGuardError) throw response.apiGuardError;

                return response;
            });
    }

    private _createRequestWithDebounce(requestConfig: DebounceRequestConfig, requests: Array<DebounceRequest<any>>) {
        axios
            .request<{ data?: Array<any>, Result?: Array<any> }>({
                ...requestConfig.axiosConfig,
                ...this._getRequestDataWithDebounce(requestConfig, requests)
            })
            .then(({ data }) => {
                (data.data || data.Result)?.forEach((responseData, index: any) => {
                    this._dispatchReponse({
                        id: requests[index].id,
                        payload: responseData
                    });
                });
            })
            .catch((e) => {
                verifyAxiosError(e);
                verifyAxiosRedirect(e);

                requests.forEach((request) => {
                    this._dispatchReponse({
                        id: request.id,
                        payload: {
                            apiGuardError: e
                        }
                    });

                });
            });
    }
    
    private _getRequestDataWithDebounce(requestConfig: DebounceRequestConfig, requests: Array<DebounceRequest>) {
        if (requestConfig.useURLSearchParams) {
            const data = new URLSearchParams();

            data.append(requestConfig.arrayPropName, JSON.stringify((requests.map((request) => request.requestData?.urlParams))))

            return { params: data };
        }

        return {
            data: { 
                [requestConfig.arrayPropName]: requests.map((request) => request.requestData?.bodyData),
            }
        };
    }

    private _dispatchReponse(reponse: DebounceResponse) {
        this._debounceResponse.next(reponse);
    }
}

export default new Api();