import { v4 as uuidv4 } from 'uuid';
import FormManager from '../../../js/FormManager';
import { Field } from '../../../js/nFormulario';
import { Field as IField, FormProp, ToolButton } from '../fra-control/types';
import { Buttons } from './buttons/buttons';
import { FormRepeaterProps } from './form-repeater';
import { FormRepeaterBase, FormRepeaterItem, RenderProps } from './form-repeater-base';
import { getFormFieldsContainer } from './form/form';
import { ItemData } from './types';
import { fixValues, splitData } from './util/uuid-map-properties/data-map';

declare var KTUtil: any;

interface FormRepeaterFieldsProps extends FormRepeaterProps {
    field: IField;
}

interface RenderPropsFields extends RenderProps {
    field: IField;
}

interface AddFieldProps {
    uuid: string;
    field: IField;
    formRepeaterItem: FormRepeaterItem;
    formProp: FormProp;
}

export interface FormRepeaterFieldsItem extends FormRepeaterItem {
    fields: Field[];
}

/**
 * The component relies on the Formulario 
 * The creation of the component must be delayed until Formulario has been instantiated, otherwise the Formulario won't be passed as property
 * 
 */
export class FormRepeaterFields extends FormRepeaterBase {
    protected config: FormRepeaterFieldsProps;

    constructor(props: FormRepeaterFieldsProps) {
        super();

        this.config = props;

        this.addInstance({
            ...props,
        });

        this.setFieldComponent(props);
        this.setScroll(this._elementRoot!);
        this.loadDataProcess(props.field);
    }

    /**
     * Must set Field component manually because formRepeater is effectively created after the ntcGrid field
     */
    private setFieldComponent(props: FormRepeaterFieldsProps) {
        const { formInstance, field } = props;

        let fieldInstance = formInstance?.fieldbyName(field.Name) as Field;

        fieldInstance.setFormRepeater(this);
        fieldInstance.componente = this._elementRoot;
        fieldInstance.componente.data('q-field', fieldInstance);
    }

    private setScroll(element: JQuery<HTMLElement>): void {
        const hasMaxHeigh = (el: JQuery<HTMLElement>): boolean => ((el.height() || 0) > 400);

        let hasScroll = false;

        element.addClass('pl-2 pr-2');

        new MutationObserver(() => {
            if (hasScroll) {
                KTUtil.scrollUpdate(element.get(0))
            }
            else if (hasMaxHeigh(element)) {
                hasScroll = true;

                KTUtil.scrollInit(element.get(0), {
                    height: () => Math.min(element.height() || 0, 400),
                });

                element.css({
                    'height': 'initial',
                    'max-height': '400px'
                })
            }
        })
        .observe(element.get(0)!, {
            childList: true
        })
    }
        

    public addInstance(props: FormRepeaterFieldsProps): FormRepeaterFieldsItem {
        let uuid = uuidv4().replace(/-/g, '_');

        let preparedItem = this.render({
            ...props,
            uuid,
            formRepeaterElementId: `ctrl${props.field.Name}`,
            buttons: this._defaultButtons(props.formProp)
        });

        let formItem = this.createFieldInstance({
            uuid,
            field: props.field,
            formRepeaterItem: preparedItem,
            formProp: props.formProp
        });

        this._items.set(uuid, formItem);

        return formItem;
    }

    private createFieldInstance(props: AddFieldProps): FormRepeaterFieldsItem {
        const { formRepeaterItem } = props;
        const { element, formComponent } = formRepeaterItem;

        let fields: Field[] = [];

        element
            .find('[data-fracontrol-script-create="true"]')
            .each((_, elementCreator) => {
                let additionalOptions = {
                    rootElement: element,
                    isFormRepeaterField: true,
                    uuid: props.uuid
                };

                let extractedOptions = FormManager.extractFormFieldOptions(formComponent, elementCreator, additionalOptions)

                extractedOptions?.forEach((option) => {
                    fields.push(formComponent.newField(option));
                })
            });

        element
            .find('[data-mask]')
            .each((_, el) => {
                // @ts-ignore
                $.applyDataMask(el)
            });

        return {
            fields,
            ...formRepeaterItem,
        };
    }

    public removeInstance(uuid: string): void {
        let instance = this._items.get(uuid);

        if (!instance) return;

        if (this._items.size > 1) {
            const { element } = instance;

            element.remove();

            this._items.delete(uuid);
        }
        else {
            const { formComponent, fieldName } = instance;

            (formComponent.fieldbyName(fieldName) as Field).clearFormRepeaterFields(uuid);
        }
    }
    protected createFormRepeaterItem(props: RenderPropsFields): FormRepeaterItem {
        const { uuid, buttons, item, formProp, field } = props;

        let formItemEl = $(`<div id="${uuid}" class="form-repeater__item d-flex d-flex-column">`);
        let col = $('<div class="d-flex w-100">');
        let { row } = getFormFieldsContainer({
            formProp,
            item,
            formInstance: props.formInstance
        });

        let buttonsEl = Buttons({
            item,
            field,
            formProp,
            uuid,
            formRepeater: this,
            toolButtons: buttons.filter(({ Name }) => (Name === 'actDataSetInsert' || Name === 'actDataSetDelete')),
            formInstance: props.formInstance
        });

        row.addClass('w-100');

        buttonsEl
            .removeClass('col')
            .addClass('d-flex')

        formItemEl
            .append(col.append(row))
            .append(buttonsEl);

        return {
            element: formItemEl,
            formComponent: props.formInstance!,
            fieldName: field.Name
        }
    }

    /**
     * 
     * Get the cofig for addButton removeButton.
     * 
     * Buttons must have the same name as an existing action (actDataSetInsert, actDataSetDelete, actDataSetPost)
     */
    protected _defaultButtons(formProp: FormProp): ToolButton[] {
        let actAdd = {
            Caption: 'Adicionar',
            Classe: formProp.Class,
            CleanCaption: 'Adicionar',
            Custom: false,
            ExibirToolBar: true,
            Hint: 'Adicionar',
            Name: 'actDataSetInsert',
            NameEvent: 'actDataSetInsertExecute',
            Opcoes: {},
            TipoAcao: 'tnAction',
            icon: '',
            tbsButton: true,
        }

        let actDelete = {
            Caption: 'Excluir',
            Classe: formProp.Class,
            CleanCaption: 'Excluir',
            Custom: false,
            ExibirToolBar: true,
            Hint: 'Excluir',
            Name: 'actDataSetDelete',
            NameEvent: 'actDataSetDeleteExecute',
            Opcoes: {},
            TipoAcao: 'tnAction',
            icon: '',
            tbsButton: true,
        }

        return [
            actAdd,
            actDelete,
        ]
    }

    public loadDataProcess(field: IField) {
        const getData = (jsonString: string = '') => {
            try {
                return JSON.parse(jsonString);
            }
            catch {
                return [];
            }
        };

        this.loadData(getData(field.Valor));
    }

    public loadData(data: ItemData[]) {
        const resetInstances = () => {
            let size = this._items.size - data.length;
            
            if (size < 1) return;

            const itens = Array.from(this._items.keys());

            for (let index = 0; index < size; index++) {
                this.removeInstance(itens[index]);
            }
        }

        resetInstances();

        const [dataToExistingInstance, dataToNewInstance] = splitData(data, this._items.size);

        const instances = Array.from(this._items.values());
        
        let { element } = instances[instances.length - 1];

        dataToExistingInstance.forEach((data, index) => {
            const fixedData = fixValues(data);

            (instances[index] as FormRepeaterFieldsItem).fields.forEach((field) => {
                field.setvalue(fixedData[field.nome]);
            });      
        });

        dataToNewInstance.forEach((data) => {
            const fixedData = fixValues(data);
            const { fields, element: newwElement } = this.addInstance({
                ...this.config,
                append: true,
                element,
            });

            fields.forEach((field) => {
                field.setvalue(fixedData[field.nome]);
            });

            element = newwElement;
        });
    }

    public getItemsArray() {
        return Array.from(this._items.values()); 
    }
}