import { FormatErrorMessage } from "../../../js/nDialog";
import SessionManager from "../../../js/SessionManager";
import { DashboardProduct, storeProductForRedirect } from "../dashboard-products/dashboard-products";
import { MenuConfig, MenuItem, ProductMenuItem, Section, SubSection } from "./menu";

export interface MenuResponse {
    ListaMenu: Array<MSection>;
    ListaFavoritos?: Array<FavoriteItem>;
    ListaRecentes?: Array<RecentItem>;
    ListaProdutos?: Array<ProductItem>;
    ListaProdutosDashboard?: Array<DashboardProduct>;
    SelecaoUnica?: boolean;
    MenuOutros?: boolean;
}

interface FavoriteItem {
    name: string;
    role: string;
    url: string;
    moduleName?: string;
    pathName?: string;
}

interface RecentItem {
    Class: string;
    TipoObjeto: string;
    Modulo?: string;
    PathName?: string;
}

export interface ProductItem {
    name: string;
    hasAccess: boolean;
    icon: string;
    url: string;
    caption: string;
    onClick?: (p: any) => void;
}

export interface MModule {
    Modulo?: string; // npm
    Childs: Array<MItem>;
    Ordem: number;
    OrdemIndex: number;
    Text: string;
}

export interface MSection {
    Childs: Array<MItem | MSubSection | MModule>;
    Ordem: number;
    OrdemIndex: number;
    Text: string;

    Colunas?: string;
    Icon?: string;
}

export interface MSubSection extends MSection { 
    title?: string // used internally
}

export interface MItem {
    Text: string;
    Ordem: number;
    OrdemIndex: number;
    Name: string;
    Tipo: string;
    ClassAreaTrabalho?: string;
    ClassWizard?: string;
    Icon?: string;
    title?: string // used internally
    onClick?: () => void;
}

interface AdditionalOptions {
    favoriteMap: FavoriteMap;
    recentMap: RecentMap;
    hasModule: boolean | undefined;
    hasTreeview?: boolean
}

type FavoriteMap = Map<string, FavoriteItem>;
type RecentMap = Map<string, RecentItem & { recentOrder: number }>;

const isMenuItem = (item: any): item is MItem => !Boolean((item as MSubSection).Childs);
const isModuleItem = (item: any): item is MModule => Boolean((item as MModule).Modulo);

export const mapConfigFromRequest = (menuResponse: MenuResponse): MenuConfig =>{
    const { ListaMenu, ListaFavoritos, ListaRecentes, ListaProdutos, SelecaoUnica, MenuOutros } = menuResponse;
    const favorites = getFavorites(ListaFavoritos);
    const recents = getRecents(ListaRecentes, SelecaoUnica);
    
    return {
        menu: {
            sections: [
                getCompanyItem(),
                ...getProductItems(menuResponse),
                ...ListaMenu.map((section) => {
                    return mapSection(
                        section,
                        {
                            favoriteMap: favorites,
                            recentMap: recents,
                            hasModule: SelecaoUnica,
                            hasTreeview: MenuOutros
                        }
                    );
                }),
            ],
            products: [], //ListaProdutos?.map(mapProducts) || [], Array vazio enquanto estiver com os produtos no dashboard
        }
    };
};

const getFavorites = (favorites: Array<FavoriteItem> | undefined) => {
    return new Map((favorites || []).map((favorite) => [
        `${padWith_(favorite.pathName)}${favorite.role}${padWith_(favorite.moduleName, true)}`.toLowerCase(),
        favorite
    ]));
};

const getRecents = (recents: Array<RecentItem> | undefined, hasModules: boolean | undefined) => {
    return new Map((recents || []).map((recent, index) => [
        `${padWith_(recent.PathName)}${recent.Class}${recent.Modulo && hasModules ? padWith_(`npm${recent.Modulo}`, true) : ''}`.toLowerCase(),
        {
            ...recent,
            recentOrder: index
        }
    ]));
};

const padWith_ = (str = '', prefix = false) => (str ? `${prefix ? '_' : ''}${str}(${prefix ? '' : '_'}` : '');

const mapSection = (section: MSection, options: AdditionalOptions): Section => {
    if (options.hasModule) {
        const items = section.Childs
            .sort((a, b) => a.OrdemIndex - b.OrdemIndex)
            .reduce((items, item) => {
                if (isModuleItem(item)) {
                    if (item?.Childs?.[0]) {
                        item.Childs[0].title = item.Text;
                    }

                    return [...items, ...(item.Childs || [])];
                }

                return [...items, item];
            }, [] as Array<MItem | MSubSection | MModule>);

        return ({
            items: items.filter(isVisible).map((child: any) => mapSubSectionOrItem(child, options)),
            caption: section.Text,
            columnSize: Number(section.Colunas),
            icon: section.Icon || getSectionItem(section),
        })
    }

    return ({
        items: section.Childs.filter(isVisible).map((child) => mapSubSectionOrItem(child, options)),
        caption: section.Text,
        columnSize: Number(section.Colunas),
        icon: section.Icon || getSectionItem(section),
    })
};

const isVisible = (subSectionOrItem: MItem | MSubSection): boolean => {
    if (isMenuItem(subSectionOrItem)) {
        return subSectionOrItem.ClassAreaTrabalho ? false : true;
    }

    return subSectionOrItem.Childs.reduce((visible, child) => (visible || isVisible(child)), false as boolean);
}

const mapSubSectionOrItem = (subSectionOrItem: MItem | MSubSection, options: AdditionalOptions) => {
    return isMenuItem(subSectionOrItem)
        ? mapItem(subSectionOrItem, options)
        : mapSubSection(subSectionOrItem, options);
};

const mapSubSection = (subSection: MSubSection, options: AdditionalOptions): SubSection => {
    if (options.hasTreeview && subSection.Text.trim().match(/^outros$/i)) {
        const isItem = (item: any): item is MItem => Boolean((item as MItem).Name);
        const isSubItem = (item: any): item is MSubSection => Boolean((item as MSubSection).Childs);
        
        const mapSubItems = (subsection: MSubSection): any => {
            const childs = subsection.Childs.filter(isVisible).map((items) => mapTreeview(items));
            const withitems = childs.filter(c => c && c.items);
            const withoutitems = childs.filter(c => c && !c.items)

            return {
                caption: subsection.Text,
                items: withitems,
                menuItems: withoutitems,
            }
        };

        const mapTreeview = (item: MItem | MSubSection | MModule) => {
            if (isItem(item)) return mapItem(item, options);
            if (isSubItem(item)) return mapSubItems(item);
        };

        return {
            caption: subSection.Text,
            title: subSection.title,
            items: mapTreeview(subSection),
            withTreeview: true,
        }
    };

    return ({
        items: subSection.Childs.filter(isVisible).map((item) => mapItem(item as any, options)),
        caption: subSection.Text,
        icon: subSection.Icon,
        title: subSection.title,
    });
}

const mapItem = (item: MItem, options: AdditionalOptions): MenuItem => ({
    caption: item.Text,
    Name: getMenuNameOptions(item, options.hasModule).menuClassName,
    moduleName: options.hasModule ? getMenuNameOptions(item, options.hasModule).moduleName : undefined,
    pathName: getMenuNameOptions(item, options.hasModule).pathName,
    href: href(item, Boolean(options.hasModule)),
    // title: `Titulo item ${Math.random()}`,
    icon: item.Icon || (item.Tipo == 'Relatorio' ? 'bg-icon-relatorio icon-default' : undefined),
    isFavorite: isFavorite(item, options.favoriteMap, options.hasModule),
    isActive: isMenuActive(item, options.hasModule),
    ...getRecentConfig(item, options.recentMap, options.hasModule)
});

export const href = (menuItem: MItem, hasModule: boolean) => {
    const escapeQuotes = (str = '') => str.replace(/"/g, '&quot;');
    const { Name, ClassAreaTrabalho, ClassWizard, Tipo } = menuItem;

    if (hasModule) {
        const { menuClassName, moduleName, pathName} = getMenuNameOptions(menuItem, hasModule);
        
        return ClassAreaTrabalho
            ? escapeQuotes(`${location.origin}/${moduleName}/Processo?Scope={Class:"${ClassAreaTrabalho}",Detalhe:"${ClassWizard ? `${ClassWizard}_` : ''}${menuClassName}",pathName:"${pathName}"}`)
            : escapeQuotes(`${location.origin}/${moduleName}/${Tipo}?Scope={Class:"${menuClassName}",pathName:"${pathName}"}`);
    }

    const { pathName, menuClassName } = getMenuNameOptions(menuItem, hasModule);

    return ClassAreaTrabalho
            ? escapeQuotes(`Processo?Scope={Class:"${ClassAreaTrabalho}",Detalhe:"${ClassWizard ? `${ClassWizard}_` : ''}${menuClassName}",pathName:"${pathName}"}`)
            : escapeQuotes(`${Tipo}?Scope={Class:"${menuClassName}",pathName:"${pathName}"}`);
};

const hasMenuInList = <K>(menuItem: MItem, map: Map<string, K>, hasModule: boolean | undefined) => Boolean(getMenuInList(menuItem, map, hasModule));

const getMenuInList = <K>(menuItem: MItem, map: Map<string, K>, hasModule: boolean | undefined) => {
    const { menuClassName, moduleName, pathName } = getMenuNameOptions(menuItem, hasModule);
    const searchMenu = (name: string) => `${padWith_(pathName)}${name}${hasModule ? padWith_(moduleName, true) : ''}`.toLowerCase();

    if (['Cadastro', 'Consulta', 'RedirectPage'].includes(menuItem.Tipo)) {
        const begin = menuClassName.slice(0, 5);
        const end = menuClassName.slice(7);

        return map.get(searchMenu(`${begin}dc${end}`)) || map.get(searchMenu(`${begin}dm${end}`));
    }
    else {
        return map.get(searchMenu(menuClassName)) || map.get(searchMenu(menuClassName));
    }
};

const isFavorite = (menuItem: MItem, mapFavorite: Map<string, FavoriteItem>, hasModule: boolean | undefined) => {
    return hasMenuInList(menuItem, mapFavorite, hasModule); // search for "menuName_moduleName" in map
};

const isMenuActive = (menuItem: MItem, hasModule: boolean | undefined) => {
    const { menuClassName, pathName } = getMenuNameOptions(menuItem, hasModule);
    const  { pageClass, pagePathName, pageWorkspaceClass} = getCurrentPageOptions();

    return (pagePathName ? pagePathName == pathName.toUpperCase() : true)
            && (pageClass === menuClassName.toUpperCase() || pageWorkspaceClass === menuClassName.toUpperCase())
};

export const getCurrentPageOptions = () => {
    const getAttr = (attr: string) => $('#SelecaoEmpresaList').attr(attr)?.toUpperCase();
    
    return {
        pageClass: getAttr('questor-class'),
        pageWorkspaceClass: getAttr('questor-class-workspace'),
        pagePathName: getAttr('questor-path-name'),
    }
};

const getRecentConfig =  (menuItem: MItem, mapFavorite: Map<string, RecentItem & { recentOrder: number }>, hasModules: boolean | undefined) => {
    const recentMenu = getMenuInList(menuItem, mapFavorite, hasModules);

    return {
        recentOrder: recentMenu?.recentOrder,
        isRecent: Boolean(recentMenu),
    }
};

const getMenuNameOptions = (menuItem: MItem, hasModule: boolean | undefined) => {
    const splits = (menuItem.Name || '').split('_'); // Ex: TnCtbDPPlPadConfigurarDRE_npmnCtb
    
    if (hasModule) {
        return {
            moduleName:  splits.pop()!,
            menuClassName: splits.pop()!,
            pathName: splits.join('_')
        }
    }

    return {
        moduleName: SessionManager.moduloAtual,
        menuClassName: splits.pop()!,
        pathName: splits.join('_'),
    }
};

const getCompanyItem = (): Section => ({
    caption: 'Home',
    icon: 'bg-icon-empresa',
    items: [{
        Name: '',
        caption: 'Home',
        href: 'Default',
        isCompany: true,
        isActive: location.pathname.includes('Default'),
    }],
    columnSize: 12,
    isCompany: true,
});

const getProductItems = (menuResponse: MenuResponse): Section[] => {
    const products: ProductItem[] = getMegaMenuProducts(menuResponse);

    if (!products) return []; 

    return [{
        caption: 'Plataforma Questor',
        isProduct: true,
        columnSize: 12,
        columns: 12,
        items: products
            // .filter(({ hasAccess }) => hasAccess)
            .map(({ icon, name, url, onClick, hasAccess }) => ({
                icon,
                caption: name,
                href: url,
                isProduct: true,
                hasAccess,
                onClick,
            } as ProductMenuItem))
    }];
};

const getMegaMenuProducts = (config: MenuResponse): ProductItem[] => {
    const { ListaProdutosDashboard, SelecaoUnica } = config;
    const quiuMenu = findMenu(config, 'TnQuiuDPAreaTrabalho');
    const products: Array<ProductItem> = [];
    const url = href(quiuMenu, Boolean(SelecaoUnica)).replace(/&quot;/g, '"');

    if (quiuMenu) {
        products.push({
            caption: 'Quiu',
            hasAccess: true,
            icon: getProductIcon('quiu'),
            name: 'Quiu',
            url: '#',
            onClick: () => {
                window.open(url, '_self');
            }
        });
    }

    ListaProdutosDashboard?.forEach(({ icon, name, url, hasAccess, title}) => {
        products.push({
            hasAccess: Boolean(hasAccess),
            icon: getProductIcon(title),
            url: '#',
            name: title,
            caption: title,
            onClick: ({ item }) => {
                const hasAccess: boolean = item.hasAccess;
                
                if (hasAccess) {
                    try {
                        storeProductForRedirect({
                            url,
                            description: '',
                            icon: item.icon,
                            title: item.caption,
                            hasAccess: item.hasAccess
                        })
                        
                        window.open('Default', '_self');
                    }
                    catch (e) {
                        FormatErrorMessage('Falha ao acessar ' + item.caption)
                    }
                }
                else {
                    window.open(url, '_blank');
                }
                
                ($("#q_mega_menu_modal") as any).modal('hide');
            },
        });
    });

    console.log('products', products);

    return products;
};

const getProductIcon = (title: string) => {
    const withSize = (icon: string) => `${icon} icon-3x`;

    if (title.match(/quiu/gi)) return withSize('bg-icon-quiu-title');
    if (title.match(/Questor Zen/ig)) return withSize('bg-icon-questor-zen-title');
    if (title.match(/hubCount/gi)) return withSize('bg-icon-hubcount-title');
    if (title.match(/Tareffa/gi)) return withSize('bg-icon-tareffa-title');
    if (title.match(/Central de Atenção Questor/gi)) return withSize('bg-icon-caq-title');

    return '';
}

const getSectionItem = ({ Text }: MSection) => {
    const text = normalize(Text);

    if (text.includes('operacoes')) return 'bg-icon-operacoes';
    if (text.includes('obrigacoes')) return 'bg-icon-obrigacoes';
    if (text.includes('plano de contas')) return 'bg-icon-plano-de-contas';

    return '';
} 

const normalize = (str: string) => (str as string & { removeDiacritics: () => string }).removeDiacritics().toLowerCase();

const mapProducts = ({ name, hasAccess, icon, url, caption }: ProductItem) => ({ name, hasAccess, icon, url, caption });

export const fixFavoritesPathName = ({ ListaMenu, ListaFavoritos, SelecaoUnica }: MenuResponse) => {
    const favorites: Array<FavoriteItem> = [];

    const validateInsertPathName = (item: MItem, className: string, array: Array<string>) => { 
        const { pathName, menuClassName } = getMenuNameOptions(item, SelecaoUnica);

        if (menuClassName == className) {
            array.push(pathName);

            return true;
        }
    };

    const findMenuPathName = (className: string) => {
        const menuPathName: Array<string> = [];

        ListaMenu.find((section) => {
            return section.Childs.find((subsectionOrItem) => {
                if (isMenuItem(subsectionOrItem)) {
                    return validateInsertPathName(subsectionOrItem, className, menuPathName)
                }
                else {
                    return subsectionOrItem.Childs.find((subItem) => {
                        if (isMenuItem(subItem)) {
                            return validateInsertPathName(subItem, className, menuPathName);
                        }
                        else {
                            return subItem.Childs.find((item) => {
                                if (isMenuItem(item)) return validateInsertPathName(item, className, menuPathName);
                            });
                        }
                    });
                }
            });
        });

        return menuPathName[0];
    };

    ListaFavoritos?.forEach((menuFav) => {
        if (menuFav.pathName) {
            favorites.push(menuFav);
        }
        else {
            const pathName = findMenuPathName(menuFav.role);

            if (pathName) {
                favorites.push({
                    ...menuFav,
                    pathName
                });
            }
        }
    });

    return favorites;
}

export const findMenu = ({ ListaMenu, SelecaoUnica}: MenuResponse, className: string) => {
    const menu: Array<MItem> = [];

    const compareMenu = (item: MItem, className: string, array: Array<MItem>) => { 
        const { menuClassName } = getMenuNameOptions(item, SelecaoUnica);

        if (menuClassName == className) {
            array.push(item);

            return true;
        }
    };

    ListaMenu.find((section) => {
        return section.Childs.find((subsectionOrItem) => {
            if (isMenuItem(subsectionOrItem)) {
                return compareMenu(subsectionOrItem, className, menu);
            }
            else {
                return subsectionOrItem.Childs.find((subItem) => {
                    if (isMenuItem(subItem)) {
                        return compareMenu(subItem, className, menu);
                    }
                    else {
                        return subItem.Childs.find((item) => {
                            if (isMenuItem(item)) return compareMenu(item, className, menu);
                        });
                    }
                });
            }
        });
    });

    return menu[0];
}