import SessionManager from "../../../js/SessionManager";
import { fixFavoritesPathName, mapConfigFromRequest, MenuResponse } from "./map-config-from-request";
import { populateQuickSearch } from '../../../js/quickSearch';
import favoriteButton from "../favorite-button/favorite-button";
import axios, { AxiosError } from "axios";
import { Dashboard } from "../dashboard-products/dashboard-products";
import { verifyAxiosError } from "../../../js/nRequests";
import dxTreeview from 'devextreme/ui/tree_view';

declare var QMenu: any;
declare var KTApp: any;
declare var $IsIntegration: undefined | boolean;
declare var $EhModal: undefined | boolean;

export interface MenuConfig {
    menu: {
        products: Array<Products>;
        sections: Array<Section>;
    }
}

export interface Products {
    name: string;
    hasAccess: boolean;
    icon: string;
    url: string;
    caption: string;
}

export interface Section {
    caption: string;
    items: Array<SubSection | MenuItem | CompanyItem | ProductMenuItem>;
    columnSize?: number;
    columns?: number;
    isCompany?: boolean;
    isProduct?: boolean;
    icon?: string;
    withTreeview?: boolean;
}

export interface SubSection {
    caption: string;
    icon?: string;
    items: Array<MenuItem>;
    title?: string;
    withTreeview?: boolean;
}

export interface MenuItem {
    caption: string;
    Name: string;
    icon?: string;
    isFavorite?: boolean;
    isActive?: boolean;
    isRecent?: boolean;
    recentOrder?: number;
    title?: string;
    href: string;
    moduleName?: string;
    pathName?: string;
    onClick?: () => void;
}

export interface CompanyItem extends MenuItem {
    isCompany: boolean;
}

export interface ProductMenuItem extends Omit<MenuItem, 'Name'> {
    isProduct?: boolean;
}

export async function CarregarMenu() {
    try {
        if (!$('#q_mega_menu').length || $EhModal || $IsIntegration) return;

        KTApp.blockPage();

        if (checkCache()) {
            const config = getCache()!;
            const { favorites, recents } = await getRecentsFavorites()

            setTimeout(() => {
                createMenu({
                    Result: {
                        ...config.menuConfig,
                        ListaFavoritos: favorites,
                        ListaRecentes: recents
                    },                                                                                                                                  
                });
            });

            return;
        }

        return axios
            .get<{ Result: MenuResponse }>('nHttpServicos/CarregarMegaMenu')
        	.then(({ data }) => {
                setCache(data.Result);

                createMenu(data);
        	});
    }
    catch (e) {
        console.error(e)
    }
    finally {
        KTApp.unblockPage();
    }
}

interface GetFavoritesResponse {
    Result: string; // Base64
}

interface GetRecentsResponse {
    Result: string; // stringJSON
}

const getRecentsFavorites = async () => {
    const [recents, favorites] = await Promise.all([
        getRecents(),
        getFavorites()
    ]);

    return {
        recents,
        favorites
    }
};

const getRecents = async () => {
    try {
        return JSON.parse((await axios.get<GetRecentsResponse>('nHttpServicos/GetRecentes')).data.Result);
    }
    catch (e) {
        console.error(e);

        verifyAxiosError(e as AxiosError);

        return [];
    }
};

const getFavorites = async () => {
    try {
        const response = atob((await axios.get<GetFavoritesResponse>('nHttpServicos/GetFavoritos')).data.Result);

        return response == '' ? [] : JSON.parse(response);
    }
    catch (e) {
        console.error(e);

        verifyAxiosError(e as AxiosError);

        return [];
    }
};

const createMenu = (data: { Result: MenuResponse }) => {
    const fixedFavorites = fixFavoritesPathName(data.Result);
    const qMenu = new QMenu({
        userName: SessionManager.userName,
        homeRoute: 'Default',
        templates: null,
        lazyLoad: true,
        useTitleConfig: data.Result.SelecaoUnica,
        productName: 'Cloud',
        ...mapConfigFromRequest({ 
            ...data.Result, 
            ListaFavoritos: fixedFavorites,
        }),
        dependencies: {
            treeview: dxTreeview,
        }
    });

    populateQuickSearch({
        listaMenu: data.Result.ListaMenu,
        hasModules: data.Result.SelecaoUnica,
        hasTreeview: data.Result.MenuOutros
    });

    favoriteButton.setCache(fixedFavorites || []);

    qMenu.on('favorite-click', ({ item }: { item: MenuItem }) => {
        favoriteButton.setFavorite({
            name: item.caption,
            role: item.Name,
            moduleName: item.moduleName,
            pathName: item.pathName,
        });
    });

    setBlockEventOnNavigation(qMenu);

    Dashboard(data.Result.ListaProdutosDashboard);
};

const setBlockEventOnNavigation = (qMenu: any) => {
    const setBlockPage = (el: JQuery<HTMLElement>) => (el || $('')).find(':not(.show-more-items) > a.navi-link:not(.btn):not([data-toggle="modal"])').attr('data-block-page-on-click', 'true');

    setBlockPage($('#q_mega_menu_modal'));
    setBlockPage(qMenu._breadcrumbMenuObject._element);
    setBlockPage(qMenu._dashboardMenuObject._element);
};

interface CachedConfig {
    menuConfig: MenuResponse;
}

const getCacheId = (moduleName?: string, hasModules?: boolean) => {
    if (SessionManager.hasModules || hasModules) {
        return 'menuQuestorCloud';
    }
    else {
        return `menu${moduleName || SessionManager.moduloAtual}`;
    }

}

const setCache = (menuConfig: MenuResponse) => {
    localStorage.setItem(getCacheId(), JSON.stringify({
        menuConfig
    }));
};

const getCache = (moduleName?: string, hasModules?: boolean) => {
    const item = localStorage.getItem(getCacheId(moduleName, hasModules));

    if (item && item !== 'undefined') return JSON.parse(item) as CachedConfig;
    
    return null;
};

const checkCache = () => {
    const cache = getCache();

    return Boolean(cache);
};

export const invalidateMenuCache = (moduleName?: string, hasModules?: boolean) => {
    const cache = getCache(moduleName, hasModules);

    if (cache) {
        localStorage.removeItem(getCacheId(moduleName, hasModules));
        sessionStorage.removeItem('itenspesquisamenu');
    }
};