import { LayoutConfig, Widget, WidgetsConfig, LayoutPageConfig, WidgetType } from "../types";
import classNames from 'classnames';
import { v4 as uuidv4 } from 'uuid';
import { Card, CardBuilder } from "./card";
import { TwFieldList } from "../tw-field-list/tw-field-list";
import { TwMasterDetail } from "../tw-master-detail/tw-master-detail";
import { Field, GroupColumn } from "../../fra-control/types";
import { mapWidgets } from "../util/map-widgets";
import { markFirstFieldList } from "../util/set-first-field-list";
import { mapFields } from "../util/map-fields";
import { ScopeConfig } from "../../../types/page-config/page-config";
import { TwParamList } from "../tw-param-list/tw-param-list";
import { TwImage } from "../tw-image/tw-image";
import { TwButton } from "../tw-button/tw-button";
import { TwImageText } from "../tw-image-text/tw-image-text";
import { TwCardDisplayer } from "../tw-card-displayer/tw-card-displayer";
import { TwImageValueDisplayer } from "../tw-image-value-displayer/tw-image-value-displayer";
import { TwLista } from "../tw-lista/tw-lista";
import { TwLabelValue } from "../tw-label-value/tw-label-value";
import { TwHomeAreaTrabalho } from "../tw-home-area-trabalho/tw-home-area-trabalho";
import { TwCxGrid } from "../tw-cx-grid/tw-cx-grid";
import { TwCardSeletor } from "../tw-card-seletor/tw-card-seletor";
import { TwArquivosTemp } from "../tw-arquivos-temp/tw-arquivos-temp";
import { TwTreeView } from "../tw-tree-view/tw-tree-view";
import { TwMsgLista } from "../tw-msg-lista/tw-msg-lista";
import { disposeTwCxGrid } from "../tw-cx-grid/dispose-tw-cx-grid";
import dxDataGrid from 'devextreme/ui/data_grid';
import { TwText } from "../tw-text/tw-text";
import { TwValueDisplayer } from "../tw-value-displayer/tw-value-displayer";
import { TwMiniCardValueDisplayer } from "../tw-mini-card-value-displayer/tw-mini-card-value-displayer";
import { TwSimple } from "../tw-simple/tw-simple";
import { FloatingMenu, FloatingMenuItems } from "./floating-menu";
import { createTwButtonControl } from "../tw-button-control/create-tw-button-control";
import widgetManager from "../widget-manager";
import { TwAgenda } from "../tw-agenda/tw-agenda";
import { TwDatasetTreeView } from "../tw-dataset-tree-view/tw-dataset-tree-view";
import { TwIframe } from "../tw-iframe/tw-iframe";

interface CommonBuildConfig {
    config: LayoutScopeConfig;
    layout: LayoutConfig;
    parent: LayoutConfig;
    card?: Card;
    isPage?: boolean;
    isSubLayout?: boolean;
    floatingMenuItems?: FloatingMenuItems;
    options?: LayoutOptions;
}

export interface LayoutOptions {
    isFrom?: {
        twMasterDetail?: boolean;
        page?: boolean;
    };
    pageOptions?: {
        toolbarBaseElement: JQuery<HTMLElement>;
        uuid?: string;
    };
    isWorkSpace?: boolean;
    isWizard?: boolean;
}

interface BuildRowConfig extends CommonBuildConfig { }

interface BuildSubLayout extends CommonBuildConfig { }

interface BuildColumnConfig extends CommonBuildConfig { }

interface BuildContentConfig extends CommonBuildConfig { }

interface BuildDataProps extends Pick<CommonBuildConfig, 'layout' | 'isPage' | 'config' | 'card' | 'options'> { 
    floatingMenuItems?: FloatingMenuItems;
}

interface BuildPagesConfig extends Omit<CommonBuildConfig, 'layout' | 'isPage'> {
    pages: LayoutPageConfig[];
}

interface BuildPageConfig extends Pick<CommonBuildConfig, 'config' | 'parent' | 'card' | 'floatingMenuItems' | 'options'>{
    page: LayoutPageConfig;
    isVisible: boolean;
    uuid: string;
    toolbarBaseElement: JQuery<HTMLElement>;
}

export interface LayoutScopeConfig extends ScopeConfig {
    mappedWidgets: Map<string, Widget>,
    fields?: Map<string, (Field | GroupColumn)>,
}

export interface InitLayoutProps {
    scope: ScopeConfig;
    card?: Card;
    element?: HTMLScriptElement;
    options?: LayoutOptions;
    floatingMenuItems?: FloatingMenuItems;
}

interface BuildLayoutProps {
    layout: LayoutConfig;
    config: LayoutScopeConfig;
    card?: Card;
    options?: LayoutOptions;
    floatingMenuItems?: FloatingMenuItems;
}

export class Layout {
    /**
     * 
     * @param widgets 
     * 
     * Search layout configuration in all positions for widget { top: ..., client: ..., fields: ...}
     */
    static getLayoutConfig(widgets: WidgetsConfig): LayoutConfig | null {
        return Object
            .values(widgets)
            .reduce((layout, position) => {
                let flayout = position.find((widget) => widget.Layout);

                if (flayout) layout = flayout.Layout!;

                return layout;
            }, null as (LayoutConfig | null));
    }

    static update(scope: ScopeConfig) {
        const { DefinicaoObjetos } = scope;
        const { Widgets } = DefinicaoObjetos;

        if (!Widgets) return;

        const mappedWidgets = mapWidgets(Widgets);
        const layout = Layout.getLayoutConfig(Widgets);
        const removeOldIfTwCxGrid = (widget: Widget) => {
            if (WidgetType.isTwCxGrid(widget)) disposeTwCxGrid(widget);
        };

        if (!layout) return;

        const missingWidgets =  Layout.getMissingWigets(mappedWidgets, layout);
        const getWidgetConfig = {
            layout,
            config: {
                ...scope,
                mappedWidgets,
            },
            options: {},
        };

        console.log('process result widget', missingWidgets)

        Array
            .from(mappedWidgets)
            .filter(([key, widget]) => !missingWidgets.has(key) && !(WidgetType.isTwFieldList(widget) || WidgetType.isTwParamList(widget)))
            .forEach(([_, widget]) => {
                const widgetInstance = widgetManager.get(widget.Name);
                
                if(widgetInstance)  {
                   Layout.updateWidgetInstance(getWidgetConfig, widget);
                }
                else {
                    removeOldIfTwCxGrid(widget);

                    const element = Layout.getWidget(getWidgetConfig, widget);
                    
                    $(`#widget_${widget.Name}`).replaceWith(element as any);
                }
            });
        
        missingWidgets.forEach((_, widgetName) => {
            const widget = mappedWidgets.get(widgetName)!;
            const element = Layout.getWidget(getWidgetConfig, widget);      
            const elementExists = $(`#widget_${widget.Name}`);
            
            removeOldIfTwCxGrid(widget);
            
            if (elementExists.length) {
                $(`#widget_${widget.Name}`).replaceWith(element as any);
            }
            else {
                const widgetColumnWrapper = $('<div class="row widget-row"><div class="col"></div></div>');

                widgetColumnWrapper.find('.col').append(element);
                
                $(`#layout_extra_row_${DefinicaoObjetos.Class}`).append(widgetColumnWrapper);
            }
        });
    }

    static initLayout({ scope, card, element, options, floatingMenuItems }: InitLayoutProps) {
        const { DefinicaoObjetos } = scope;
        const { Widgets } = DefinicaoObjetos;

        if (!Widgets) return;

        const mappedWidgets = mapWidgets(Widgets);
        const layout = Layout.getLayoutConfig(Widgets);
        
        if (!layout) throw new Error('No layout found');

        markFirstFieldList(mappedWidgets, layout);

        console.log('layout', layout);
        console.log('mappedWidgets', mappedWidgets);

        if (DefinicaoObjetos.TipoObjeto === 'ntoProcesso') { // create new row at end to put missing widgets
            layout.Linhas?.push({
                Tamanho: 0,
                TamanhoMinimo: 0,
                TipoTamanho: 0,
                WidgetName: `layout_extra_row_${DefinicaoObjetos.Class}`,
            });
        }

        const layoutContent = Layout.buildLayout({
            layout,
            card,
            options,
            floatingMenuItems,
            config: {
                ...scope,
                mappedWidgets,
                fields: mapFields(DefinicaoObjetos.Fields, DefinicaoObjetos.Pages),
            },
        });

        if (element) {
            $(element).replaceWith(layoutContent);
        }
        else {
            return layoutContent;
        }
    }

    static buildLayout({ config, layout, card, options, floatingMenuItems }: BuildLayoutProps): JQuery<HTMLElement> {
        const { DefinicaoObjetos: { TipoObjeto }} = config;
        const { Linhas } = layout;
        const renderFloatingMenu = ((Linhas?.length ?? 0) + (TipoObjeto === 'ntoProcesso' ? 0 : 1)) > 2;
        const wrapper = $(`<div class="container-fluid layout-component ${options?.isWorkSpace || options?.isWizard ? 'p-0' : ''}">`);
        const props: BuildContentConfig = {
            config,
            layout,
            card,
            parent: layout,
            options,
            floatingMenuItems: floatingMenuItems ? floatingMenuItems : (renderFloatingMenu && !options?.isFrom?.twMasterDetail) ? [] : undefined,
        };
        const content = Layout.buildContent(props);

        if (renderFloatingMenu && ((props?.floatingMenuItems?.length || 0) > 2)) {
            wrapper.append(FloatingMenu({
                items: props.floatingMenuItems!,
                title: config.DefinicaoObjetos.Caption,
            }));
        }

        return wrapper.append(content);
    }

    static getMissingWigets(mappedWidgets: Map<string, Widget>, layout: LayoutConfig) {
        const missing = new Map<string, boolean>();
        
        mappedWidgets.forEach((_, key) => missing.set(key, true));

        const mark = (layout: LayoutConfig) => {
            if (layout.WidgetName) {
                missing.delete(layout.WidgetName);
            }
            else if (layout.Linhas) {
                for (let row of layout.Linhas) {
                    mark(row);
                }
            }
            else if (layout.Colunas) {
                for (let col of layout.Colunas) {
                    mark(col);
                }
            }
            else if (layout.Pages) {
                for (let page of layout.Pages) {
                    mark(page.Coluna);
                }
            }
            else if (layout.SubLayout) {
                mark(layout.SubLayout);
            }
        }

        mark(layout);

        return missing;
    }

    static buildContent(props: BuildContentConfig): JQuery<HTMLElement> | JQuery<HTMLElement>[] {
        const { layout, isPage, isSubLayout, config, card, floatingMenuItems, options } = props;

        if (layout.Linhas) {
            return layout.Linhas.map((row) => Layout.buildLinha({
                isPage,
                isSubLayout,
                layout: row,
                parent: layout,
                config,
                card,
                floatingMenuItems,
                options,
            }));
        }

        if (layout.Colunas) {
            return layout.Colunas.map((column) => Layout.buildColuna({
                isPage,
                isSubLayout,
                layout: column,
                parent: layout,
                config,
                card,
                floatingMenuItems,
                options,
            }));
        }

        if (layout.Pages) {
            return Layout.buildPages({
                pages: layout.Pages,
                parent: layout,
                config,
                card,
                floatingMenuItems,
                options,
            });
        }

        if (layout.SubLayout) {
            return Layout.buildSubLayout({
                layout: layout.SubLayout,
                isPage,
                isSubLayout,
                parent: layout,
                config,
                card,
                floatingMenuItems,
                options,
            });
        }

        return Layout.buildData({ config, layout, isPage, card, floatingMenuItems, options });
    }

    static buildLinha(opt: BuildRowConfig): JQuery<HTMLElement> {
        const { config, layout, parent, isPage, isSubLayout, card, floatingMenuItems, options } = opt;
        const row = $('<div>').addClass('row');
        const content = Layout.buildContent({
            config,
            layout,
            parent,
            card,
            isPage,
            isSubLayout,
            floatingMenuItems,
            options,
        });

        return row.append(content);
    }

    static buildColuna(props: BuildColumnConfig) {
        const { layout, isPage, isSubLayout, parent, config, card, floatingMenuItems, options } = props;
        const { TipoTamanho, Tamanho } = layout;
        
        const column = $(`<div class="col">`).css('max-width', `${Tamanho}${TipoTamanho === 0 ? '%' : 'px'}`);

        const buildContent = (card: Card) => {
            return Layout.buildContent({
                isPage,
                isSubLayout,
                layout,
                parent,
                config,
                card: card,
                floatingMenuItems,
                options,
            })
        }
        
        if (card) {
            const content = buildContent(card);

            return column.append(content);
        }
        else {
            const card = CardBuilder.build({
                title: layout.Titulo,
                showHeader: layout.Pages  ? false : true,
            });

            const content = buildContent(card);
            
            if (!layout.Pages && layout.Titulo) {
                floatingMenuItems?.push({
                    caption: layout.Titulo,
                    el: card.cardElement,
                });
            }

            CardBuilder.setContent(card, content);
            
            return column.append(card.cardElement);
        }
    }

    static buildPages(props: BuildPagesConfig) {
        const { pages, parent, config, card, floatingMenuItems, options } = props;

        const buildLink = (page: LayoutPageConfig, active: boolean, id: string) => {
            const tabLinks = $(
                `<li class="nav-item">
                    <a 
                        class="nav-link ${active ? 'active' : ''}" 
                        data-toggle="tab" 
                        href="#tab_${id}"
                    >${page.Caption}</a>
                </li>`
            );

            tabLinks
                .find('a')
                .on('shown.bs.tab', ({ currentTarget }) => { // repaint grids on tab change
                    const tabId = (currentTarget).href.match(/#tab_(\S+)/)?.[1];
                    const tab = $(`#tab_${tabId}`);

                    console.log('tabid', tabId, $(tab));

                    $(`[id*="dataGridConteudo"]:visible`, tab)
                        .toArray()
                        .forEach((gridEl) => {
                            try {
                                const gridComponent = dxDataGrid.getInstance(gridEl) as dxDataGrid;

                                gridComponent.repaint();
                                gridComponent.refresh();
                            }
                            catch (e) {
                                console.error(e);
                            }
                        });
                });

            return tabLinks;
        };
        const setToolbutonEvents = (ul: JQuery<HTMLElement>) => {
            const tabLinks = ul.find('li.nav-item');
            
            tabLinks
                .find('a')
                .on('shown.bs.tab', ({ currentTarget }) => { // repaint grids on tab change
                    const tabId = (currentTarget).href.match(/#tab_(\S+)/)?.[1];
                    const toolbutonsTab = $(`#toolbuttons_tab_${tabId}`, ul);

                    toolbutonsTab.show();
                })
                .on('hide.bs.tab', ({ currentTarget }) => {
                    const tabId = (currentTarget).href.match(/#tab_(\S+)/)?.[1];
                    const toolbutonsTab = $(`#toolbuttons_tab_${tabId}`, ul);

                    toolbutonsTab.hide();
                });
        }

        const ul = $('<ul class="nav nav-tabs nav-tabs-line">');
        const liButtons = $('<li class="ml-auto">');
        const tabs = $('<div class="tab-content mt-5">');

        pages.forEach((page, index) => {
            let uuid = uuidv4();

            ul.append(buildLink(page, !index, uuid));

            const tab = Layout.buildPage({
                page,
                uuid,
                isVisible: !index,
                parent,
                config,
                card,
                floatingMenuItems,
                options,
                toolbarBaseElement: liButtons,
            })

            tabs.append(tab);
            
            if (!index) {
                $(`#toolbuttons_tab_${uuid}`, ul).show();
            }
        });

        ul.append(liButtons);

        setToolbutonEvents(ul);

        return [ul, tabs];
    }

    static buildPage(props: BuildPageConfig): JQuery<HTMLElement> {
        const { page, isVisible, uuid, parent, config, card, floatingMenuItems, options, toolbarBaseElement } = props;

        const tab = $('<div>')
            .attr({
                id: `tab_${uuid}`,
                role: 'tabpanel',
                class: classNames({
                    'tab-pane fade': true,
                    'show active': isVisible,
                }),
            });

        floatingMenuItems?.push({
            caption: page.Caption,
            el: tab,
        });

        const opt = {
            ...options,
            isFrom: {
                page: true,
            },
            pageOptions: {
                toolbarBaseElement: toolbarBaseElement,
                uuid,
            },
        };

        const content = Layout.buildColuna({
            layout: page.Coluna,
            isPage: true,
            parent,
            config,
            card,
            floatingMenuItems,
            options: opt,
        });

        return tab.append(content);
    }

    static buildSubLayout(props: BuildSubLayout) {
        const { isPage, layout, parent, config,  card, floatingMenuItems, options } = props;

        const content = Layout.buildContent({
            layout: layout,
            isPage,
            isSubLayout: true,
            parent,
            config,
            card,
            floatingMenuItems,
            options,
        });

        return content;
    }

    static buildData(props: BuildDataProps) {
        const { layout, config } = props;
        const widget = config.mappedWidgets.get(layout.WidgetName!);

        return Layout.getWidget(props, widget);
    }

    static getWidget({ layout, isPage, config, card, floatingMenuItems, options }: BuildDataProps, widget?: Widget) {
        if (!widget) {
            return layout.WidgetName == `layout_extra_row_${config.DefinicaoObjetos.Class}` ? $(`<div id="layout_extra_row_${config.DefinicaoObjetos.Class}" class="col">`) : [];
        }

        if (!Layout.canBuildVisibleWidget(widget)) return $(`<div id="widget_${widget.Name}" class="hide">`);

        if (WidgetType.isTwMasterDetail(widget)) {
            return TwMasterDetail({
                widget,
                detailsConfig: config.DefinicaoObjetos.DefinicaoDetalhes!,
                card,
                floatingMenuItems,
                options,
            });
        }

        if (WidgetType.isTwFieldList(widget)) {
            return TwFieldList({
                widget,
                config,
                card: card!,
                floatingMenuItems,
                options,
            });
        }

        if (WidgetType.isTwParamList(widget)) {
            return TwParamList({
                config,
                widget,
                card: card!,
                floatingMenuItems,
                options,
            });
        }

        if (WidgetType.isTwImage(widget)) {
            return TwImage({ widget });
        }

        if (WidgetType.isTwButton(widget)) {
            return TwButton({ 
                widget,
                DefinicaoObjetos: config.DefinicaoObjetos,
            });
        }

        if (WidgetType.isTwImageText(widget)) {
            return TwImageText({ widget });
        }

        if (WidgetType.isTwCardDisplayer(widget)) {
            return new TwCardDisplayer({
                widget,
                Class: config.DefinicaoObjetos.Class,
                isLayout: true,
            }).element;

        }

        if (WidgetType.isTwImageValueDisplayer(widget)) {
            return new TwImageValueDisplayer({
                widget,
                Class: config.DefinicaoObjetos.Class,
            }).element;
        }

        if (WidgetType.isTwLista(widget)) {
            return TwLista({ widget });
        }

        if (WidgetType.isTwLabelValue(widget)) {
            return TwLabelValue({
                widget,
                DefinicaoObjetos: config.DefinicaoObjetos,
            });
        }

        if (WidgetType.isTwHomeAreaTrabalho(widget)) {
            return TwHomeAreaTrabalho({ widget });
        }

        if (WidgetType.isTwCxGrid(widget)) {
            return TwCxGrid({
                widget,
                DefinicaoObjetos: config.DefinicaoObjetos,
                Params: config.Params,
            });
        }

        if (WidgetType.isTwCardSeletor(widget)) {
            return new TwCardSeletor({ 
                widget,
                DefinicaoObjetos: config.DefinicaoObjetos,
            }).element;
        }
        

        if (WidgetType.isTwArquivosTemp(widget)) {
            return TwArquivosTemp({ widget });
        }

        if (WidgetType.isTwTreeView(widget)) {
            return TwTreeView({ widget });
        }

        if (WidgetType.isTwMsgLista(widget)) {
            return TwMsgLista({ widget });
        }

        if (WidgetType.isTwText(widget)) {
            return TwText({ widget });
        }

        if (WidgetType.isTwValueDisplayer(widget)) {
            return new TwValueDisplayer({ 
                widget,
                Class: config.DefinicaoObjetos.Class,
            }).element;
        }

        if (WidgetType.isTwMiniCardValueDisplayer(widget)) {
            return new TwMiniCardValueDisplayer({ 
                widget, 
                isLayout: true,
                Class: config.DefinicaoObjetos.Class,
            }).element;
        }

        if (WidgetType.isTwSimple(widget)) {
            return TwSimple({ widget });
        }

        if (WidgetType.isTwButtonControl(widget)) {
            const element = $('<div>');

            createTwButtonControl({
                widget,
                element: element,
                DefinicaoObjetos: config.DefinicaoObjetos,
            });

            return element;
        }

        if (WidgetType.isTwAgenda(widget)) {
            return new TwAgenda({ 
                widget,
                DefinicaoObjetos: config.DefinicaoObjetos,
            }).element;
        }

        if (WidgetType.isTwDatasetTreeView(widget)) {
            return new TwDatasetTreeView({ widget }).element;
        }

        if (WidgetType.isTwIframe(widget)) {
            return TwIframe({ widget: widget });
        }

        if (isPage) {
            return $(`<label>Content Example ${layout.WidgetName} ${Math.random()}</label>`);
        }


        if (layout.WidgetName) return $(`<div>widget ${layout.WidgetName} ${Math.random()}</div>`)

        return $();
    };

    static canBuildVisibleWidget(widget?: Widget) {
        return widget
            && (
                widget.Visible
                || WidgetType.isTwParamList(widget)
                || WidgetType.isTwFieldList(widget)
            );
    }

static updateWidgetInstance({/* layout, isPage, config, card, floatingMenuItems, options*/ }: BuildDataProps, widget?: Widget) {
        if (!widget) return;

        if (WidgetType.isTwValueDisplayer(widget)) {
            widgetManager
                .get<TwValueDisplayer>(widget.Name)
                ?.setState({ widget });
        }
        else if (WidgetType.isTwImageValueDisplayer(widget)) {
            widgetManager
                .get<TwImageValueDisplayer>(widget.Name)
                ?.setState({ widget });
        }
        else if (WidgetType.isTwMiniCardValueDisplayer(widget)) {
            widgetManager
                .get<TwMiniCardValueDisplayer>(widget.Name)
                ?.setState({ widget, isLayout: true });
        }
        else if (WidgetType.isTwValueDisplayer(widget)) {
            widgetManager
                .get<TwValueDisplayer>(widget.Name)
                ?.setState({ widget });
        }
        else if (WidgetType.isTwAgenda(widget)) {
            widgetManager
                .get<TwAgenda>(widget.Name)
                ?.setState({ widget });
        }
        else if (WidgetType.isTwDatasetTreeView(widget)) {
            widgetManager
                .get<TwDatasetTreeView>(widget.Name)
                ?.setState({ widget });
        }
        else {
            throw new Error('Widget Update not implemented')
        }
    }
}
