import moment from "moment";
import FormManager from "../../../../js/FormManager";
import widgetManager from "../widget-manager";
import { Formulario } from "../../../../js/nFormulario";
import { DefinicaoObjetosConfig } from "../../../types/page-config/page-config";
import { ToolButton } from "../../fra-control/types";
import { getToolButtonAction } from "../../toolbuttons/toolbutton-action/toolbutton-action";
import { TwButtonActive, TwButtonActiveConfig } from "../tw-button-active/tw-button-active";
import { createTwButtonControl } from "../tw-button-control/create-tw-button-control";
import { TwCalendario } from "../tw-calendario/tw-calendario";
import { createTwImageButton } from "../tw-image-button/create-tw-image-button";
import { TwButtonControlConfig, Widget, WidgetConfigBase, WidgetsConfigItem, WidgetType } from "../types";
import { WidgetBase } from "../widget-base";
import { Subject } from "rxjs";

export interface TwAgendaConfig extends WidgetConfigBase {
    ContainerLeft: WidgetsConfigItem;
    ContainerClient: WidgetsConfigItem;
    ContainerRight: WidgetsConfigItem;
    colunas?: number;
    OnUpdate: ToolButton;
    Action: ToolButton;
    DataAtual: string;
    DataFinal: string;
    DataInicial: string;
    Filtro: 0 | 1 | 2; //'faDia' | 'faSemana' | 'faMes';
    twAgenda: true;
}

interface TwAgendaProps {
    widget: TwAgendaConfig;
    DefinicaoObjetos: Pick<DefinicaoObjetosConfig, 'Class' | 'TipoObjeto'>;
}

interface TwAgendaFilterParams {
    pWidgetAgendaFiltroDiaSemanaMes: string;
    pWidgetAgendaDataIni: string;
    pWidgetAgendaDataFim: string;
    pWidgetAgendaStatus: string;
    [key: string]: string;
}

type TwAgendaButtonsName =
    | 'Agendar'
    | 'Calendario'
    | 'FiltrarAnterior'
    | 'FiltrarProximo'
    | 'FiltroHoje'
    | 'FiltrarDia'
    | 'FiltrarSemana'
    | 'FiltrarMes';

type TwAgendaMomentTypeFilter = 'day' | 'month' | 'week';

const getMomentFilterType = (typeNum: number) => ['day', 'week', 'month'][typeNum] as TwAgendaMomentTypeFilter;
const getNumberFilterType = (type: TwAgendaMomentTypeFilter) => ({ day: '0', week: '1', month: '2' })[type];

export class TwAgenda extends WidgetBase<TwAgendaConfig, TwAgendaProps> {
    private _calendar: TwCalendario | undefined;
    private _params: TwAgendaFilterParams | undefined;
    private _dateFormatForRequestParam = 'YYYY-MM-DD';
    private _onFilter: Subject<TwAgendaFilterParams> | undefined;

    constructor(config: TwAgendaProps) {
        super(config);

        widgetManager.set(config.widget.Name, this);
    }

    protected _build(): JQuery<HTMLElement> {
        const compareAlignment = (a: Widget, b: Widget) => {
            const toNumber = (str?: string) => str == 'twaRight' ? 1 : str == 'twaLeft' ? -1 : 0;

            return toNumber(a.Align) - toNumber(b.Align);
        };

        const { widget } = this.getState();
        const { ContainerClient, ContainerLeft, ContainerRight } = widget;
        const getWidgets = (widgets: Widget[]) => widgets.map((widget) => this._getWidget(widget));
        const left = getWidgets(ContainerLeft.Itens.sort(compareAlignment));
        const right = getWidgets(ContainerRight.Itens.sort(compareAlignment));
        const middleLeft = getWidgets(ContainerClient.Itens.filter((a) => a.Align != 'twaRight'));
        const middleRight = getWidgets(ContainerClient.Itens.filter((a) => a.Align == 'twaRight'));

        const wrapper = $(/*html*/`
            <div class="d-flex ${widget.colunas ? `col-${widget.colunas}` : 'flex-grow-1'}">
                <div class="d-flex tw-agenda-left align-items-center"></div>
                <div class="d-flex flex-grow-1 tw-agenda-middle align-items-center">
                    <div class="d-flex flex-grow-1 tw-agenda-middle-left"></div>
                    <div class="d-flex flex-grow-1 tw-agenda-middle-right justify-content-end"></div>
                </div>
                <div class="d-flex tw-agenda-right ml-auto align-items-center"></div>
            </div>
        `);

        wrapper.find('.tw-agenda-left').append(left);
        wrapper.find('.tw-agenda-middle-left').append(middleLeft);
        wrapper.find('.tw-agenda-middle-right').append(middleRight);
        wrapper.find('.tw-agenda-right').append(right);

        this._initEvents();
        this._initFilter();

        return wrapper;
    }

    private _getWidget(widgetItem: Widget) {
        switch (widgetItem.Name as TwAgendaButtonsName) {
            case 'Agendar': return this._buttonAgendaAction(widgetItem); // twButtonControl
            case 'FiltrarAnterior': return this._buttonFilterPrevious(widgetItem); // twImageButton
            case 'FiltrarDia': return this._buttonFilterDay(widgetItem as TwButtonActiveConfig); // twButtonActive
            case 'FiltrarMes': return this._buttonFilterMonth(widgetItem as TwButtonActiveConfig); // twButtonActive
            case 'FiltrarSemana': return this._buttonFilterWeek(widgetItem as TwButtonActiveConfig); // twButtonActive
            case 'FiltroHoje': return this._buttonFilterToday(widgetItem); // twButtonControl
            case 'FiltrarProximo': return this._buttonFilterNext(widgetItem); // twImageButton
        }

        return this._widgetBuilder(widgetItem);
    }

    private _buttonFilterPrevious(widgetItem: Widget) {
        const widgetEl = this._widgetBuilder(widgetItem);

        widgetEl.on('click', () => {
            const type = getMomentFilterType(Number(this._params!.pWidgetAgendaFiltroDiaSemanaMes));
            const startDate = moment(this._calendar?.getCurrentDate()).subtract(1, type)

            if (this._calendar) {
                this._calendar.setDate(startDate.toDate());
            }
            else {
                this._filter({
                    ...this._params!,
                    pWidgetAgendaDataIni: this._getStartDate(startDate, type),
                    pWidgetAgendaDataFim: this._getEndDate(startDate, type),
                });
            }
        });

        return widgetEl;
    }

    private _buttonFilterNext(widgetItem: Widget) {
        const widgetEl = this._widgetBuilder(widgetItem);

        widgetEl.on('click', () => {
            const type = getMomentFilterType(Number(this._params!.pWidgetAgendaFiltroDiaSemanaMes));
            const date = moment(this._calendar?.getCurrentDate()).add(1, type);

            if (this._calendar) {
                this._calendar.setDate(date.toDate());
            }
            else {
                this._filter({
                    ...this._params!,
                    pWidgetAgendaDataIni: this._getStartDate(date, type),
                    pWidgetAgendaDataFim: this._getEndDate(date, type),
                });
            }
        });

        return widgetEl;
    }

    private _buttonFilterToday(widgetItem: Widget) {
        const widgetEl = this._widgetBuilder(widgetItem);

        widgetEl.on('click', () => {
            const type = getMomentFilterType(Number(this._params!.pWidgetAgendaFiltroDiaSemanaMes));
            const date = moment();

            if (this._calendar) {
                this._calendar.setDate(date.toDate());
            }
            else {
                this._filter({
                    ...this._params!,
                    pWidgetAgendaDataIni: this._getStartDate(date, type),
                    pWidgetAgendaDataFim: this._getStartDate(date, type),
                });                
            }
        });

        return widgetEl;
    }

    private _buttonAgendaAction(widgetItem: Widget) {
        const { widget } = this.getState();
        const { Classe } = widget.Action;

        const widgetEl = this._widgetBuilder({
            ...widgetItem,
            Action: widget.Action,
        } as TwButtonControlConfig);


        FormManager.runAfterCreateAllForms(() => {
            const form: Formulario = FormManager.byname(Classe);
            const afterModal = form.AfterModal;

            form.AfterModal = (...args: any) => {
                if (afterModal) {
                    afterModal(...args);
                }

                this._filter({});
            }
        });

        return widgetEl;
    }

    private _buttonFilterDay(widgetItem: TwButtonActiveConfig) {
        const widgetEl = this._widgetBuilder(widgetItem);

        widgetEl.on('click', () => this._changeFilterInterval('day'));

        return widgetEl;
    }

    private _buttonFilterWeek(widgetItem: TwButtonActiveConfig) {
        const widgetEl = this._widgetBuilder(widgetItem);

        widgetEl.on('click', () => this._changeFilterInterval('week'));

        return widgetEl;
    }

    private _buttonFilterMonth(widgetItem: TwButtonActiveConfig) {
        const widgetEl = this._widgetBuilder(widgetItem);

        widgetEl.on('click', () => this._changeFilterInterval('month'));

        return widgetEl;
    }

    private _changeFilterInterval(intervalType: TwAgendaMomentTypeFilter) {
        const currentDate = moment(this._calendar?.getCurrentDate());

        this._calendar?.setIntervalType(intervalType);

        this._filter({
            ...this._params!,
            pWidgetAgendaFiltroDiaSemanaMes: getNumberFilterType(intervalType),
            pWidgetAgendaDataIni: this._getStartDate(currentDate, intervalType),
            pWidgetAgendaDataFim: this._getEndDate(currentDate, intervalType),
        });
    }

    private _widgetBuilder(widgetItem: Widget) {
        const container = $(`<div class="px-1">`);
        const rootEl = $('<div>');

        if (WidgetType.isTwImageButton(widgetItem)) {
            createTwImageButton({
                scope: widgetItem,
                element: rootEl,
            });

            return container.append(rootEl);
        }
        else if (WidgetType.isTwButtonControl(widgetItem)) {
            const { DefinicaoObjetos } = this.getState();
            createTwButtonControl({
                DefinicaoObjetos,
                element: rootEl,
                widget: widgetItem,
            });

            return container.append(rootEl);
        }
        else if (WidgetType.isTwCalendario(widgetItem)) {
            if (this._calendar) {
                this._calendar.setState({ widget: widgetItem });

                return this._calendar.element;
            }

            this._calendar = new TwCalendario({ widget: widgetItem });

            this._listenForCalendarChanges();

            return container.append(this._calendar.element);
        }
        else if (WidgetType.isTwButtonActive(widgetItem)) {
            return container.append(TwButtonActive({ widget: widgetItem }));
        }

        return $();
    }

    private _getEndDate(date: moment.Moment, intervalType?: moment.unitOfTime.StartOf) {
        const { widget } = this.getState();

        return date.endOf(intervalType || getMomentFilterType(widget.Filtro)).format(this._dateFormatForRequestParam);
    }

    private _getStartDate(date: moment.Moment, intervalType?: moment.unitOfTime.StartOf) {
        const { widget } = this.getState();

        return date.startOf(intervalType || getMomentFilterType(widget.Filtro)).format(this._dateFormatForRequestParam);
    }

    private _initFilter() {
        const { widget } = this.getState();
        const currentDate = moment(this._calendar?.getCurrentDate());

        this._params = {
            pWidgetAgendaDataFim: this._getEndDate(currentDate),
            pWidgetAgendaDataIni: currentDate.format(this._dateFormatForRequestParam),
            pWidgetAgendaFiltroDiaSemanaMes: `${widget.Filtro}`,
            pWidgetAgendaStatus: '0',
        };
    }

    private _initEvents() {
        this._onFilter = new Subject<TwAgendaFilterParams>();
    }

    private _listenForCalendarChanges() {
        this._calendar?.OnChangeDate().subscribe((newDate) => {
            const date = moment(newDate);
            const type = getMomentFilterType(Number(this._params!.pWidgetAgendaFiltroDiaSemanaMes));

            this._filter({
                ...this._params!,
                pWidgetAgendaDataIni: this._getStartDate(date, type),
                pWidgetAgendaDataFim: this._getEndDate(date, type),
            });
        });
    }

    private _filter(Params: Partial<TwAgendaFilterParams>) {
        const { widget: { OnUpdate }, DefinicaoObjetos } = this.getState();

        this._setParams(Params);

        this._onFilter?.next(this._params);

        getToolButtonAction({
            Params: this._params,
            DefinicaoObjetos,
            toolButton: {
                ...OnUpdate,
                Opcoes: {
                    ...OnUpdate.Opcoes,
                    DataSend: 'dsNone',
                }
            },
        })();
    }

    private _setParams(params: Partial<TwAgendaFilterParams>) {
        this._params = {
            ...this._params,
            ...(params as TwAgendaFilterParams),
        };

    }

    public filter(Params: Partial<TwAgendaFilterParams>) {
        this._filter(Params);
    }

    public getParams() {
        return this._params;
    }

    public onFilter() {
        return this._onFilter;
    }
}
