import dxPopup from 'devextreme/ui/popup';
import dxCalendar from 'devextreme/ui/calendar';
import { WidgetBase } from "../widget-base";
import { WidgetConfigBase } from '../types';
import moment from 'moment';
import { Subject } from 'rxjs';

export interface TwCalendarioConfig extends WidgetConfigBase { 
    twCalendario: true;
}

interface TwCalendarioProps {
    widget: TwCalendarioConfig,
}

type IntervalType = 'day' | 'month' | 'week';

export class TwCalendario extends WidgetBase<TwCalendarioConfig, TwCalendarioProps> {
    private _popup: dxPopup | undefined;
    private _calendar: dxCalendar | undefined;
    private _previousDate = new Date();
    private _currentDate = new Date();
    private _intervalType: IntervalType = 'day';
    private _onChangeDate = new Subject<Date>();

    constructor(config: TwCalendarioProps) {
        super(config);

    }

    protected _build(): JQuery<HTMLElement> {
        this._element = $(`<div><span class="btn btn-text-dark btn-hover-bg-light">${this._dateToText(new Date())}</span></div>`);

        if (!this._calendar) this._calendar = this._createCalendar();
        if (!this._popup) this._popup = this._createPopup();

        this._element.on('click', () => {
            this._popup?.show();
        })

        this.element.append(this._popup.element());

        return this._element;
    }

    private _createCalendar() {
        return new dxCalendar(document.createElement('div'), {
            onValueChanged: ({ value, previousValue }) => {
                this._previousDate = previousValue || new Date();
                this._currentDate = value;

                this._onChangeDate.next(value);

                this._updateDate(value as Date);
            },
            activeStateEnabled: true
        });
    }

    private _createPopup() {
        const el = document.createElement('span');

        return new dxPopup(el, {
            contentTemplate: () => this._calendar?.element(),
            position: {
                of: el,
                at: 'bottom'
            },
            title: 'Calendário',
            width: 'auto',
            height: 'auto',
            shading: false,
            resizeEnabled: false,
            dragEnabled: false,
            closeOnOutsideClick: true,
            toolbarItems: [
                {
                    toolbar: 'bottom',
                    location: 'after',
                    template: () => {
                        return $(/*html*/`<button class="btn btn-sm btn-q-secondary"><span class="h4">CANCELAR</span></button>`)
                            .on('click', () => {
                                this._calendar?.option('value', this._previousDate);
                                this._popup?.hide();
                            });
                    },
                },
                {
                    toolbar: 'bottom',
                    location: 'after',
                    template: () => {
                        return $(/*html*/`<button class="btn btn-sm btn-q-primary"><span class="h4">CONFIRMAR</span></button>`)
                            .on('click', () => {
                                this._popup?.hide();
                            });
                    },
                }
            ],
        });
    }

    private _dateToText(date: Date) {
        switch(this._intervalType) {
            case 'week': 
                const startDate = moment(date).startOf('week');
                const endDate = moment(date).endOf('week');
                const isSameMonth = startDate.isSame(endDate, 'month');
                const startStr = startDate.toDate().toLocaleDateString('pt-br', { day: '2-digit', month: isSameMonth ? undefined : 'short' });
                const endStr = endDate.toDate().toLocaleDateString('pt-br', { day: '2-digit',  year: 'numeric', month: isSameMonth ? 'short' : 'long' });
                
                return `${startStr} - ${endStr}`;
            case 'month': return date.toLocaleDateString('pt-br', { year: 'numeric', month: 'long'});
            default: return date.toLocaleDateString('pt-br', { year: 'numeric', month: 'long', day: '2-digit' });
        }
    }

    private _updateDate(date: Date) {
        this._element.find('span.btn').text(this._dateToText(date));
    }

    private _setIntervalType(type: IntervalType) {
        this._intervalType = type;

        this._updateDate(this._currentDate);
    }

    public dispose() {
        this._calendar?.dispose();
        this._popup?.dispose();
        this._element.remove();
    }
    
    public getCurrentDate() {
        return this._currentDate;
    }

    public setDate(date: Date) {
        this._calendar?.option('value', date);
    }

    public setIntervalType(type: IntervalType) {
        this._setIntervalType(type);
    }

    public OnChangeDate() {
        return this._onChangeDate;
    }
}