import dxTreeView from 'devextreme/ui/tree_view';
import { WidgetBase } from "../widget-base";
import { WidgetConfigBase } from '../types';
import widgetManager from "../widget-manager";
import gridManager from '../../grid-manager/grid-manager';
import { GridActionHelper } from '../../../../js/nGrid';
import broadcastManager, { SaveDataType } from '../../../utils/broadcast-manager/broadcast-manager';
import { filter } from 'rxjs/operators';
import { Column } from 'devextreme/ui/data_grid';
import moment from 'moment';
import DevExpress from 'devextreme/bundles/dx.all';
import data_source from 'devextreme/data/data_source';
import array_store from 'devextreme/data/array_store';
import { Subscription } from 'rxjs';

export interface TwDatasetTreeViewConfig extends WidgetConfigBase {
    Itens: Array<Itens>;
    Titulo: string;
    twDatasetTreeView: true;
    colunas?: number;
}

interface Itens {
    ExpressaoTitulo: string; // FieldTitutlo
    DataSet: string; // nomeGrid
    DetailFields: string; // field1;field2
    MasterFields: string; // field1;field2
    ClasseCadastro: string;
}

export interface TwDatasetTreeViewProps {
    widget: TwDatasetTreeViewConfig,
}

interface RowData {
    [key: string]: any;
}

type ItemDependecyMap = Map<string, Map<string, RowData | RowData[]>>;

export class TwDatasetTreeView extends WidgetBase<TwDatasetTreeViewConfig, TwDatasetTreeViewProps> {
    private _treeview: dxTreeView | undefined;
    private _gridData = new Map<string, Array<{ [key: string]: any }>>();
    private _dataSource: data_source | undefined;
    private _expandedNodes: any[] = [];
    private _subscriptions: Array<Subscription> = [];
    private _broadCastSubscription: Subscription;

    constructor(config: TwDatasetTreeViewProps) {
        super(config);

        widgetManager.set(config.widget.Name, this);

        this._broadCastSubscription = this._listenForMessages();
    }

    private _listenForMessages() {
        const formatTypeForGrid = ({ dataType }: Column, value: any) => {
            if ([undefined, null, ''].includes(value)) return '';

            switch(dataType) {
                case 'date': return moment(value).format('MM/DD/YYYY');
                case 'number': return value.replaceAll('.', '').replaceAll(',', '.');
                default: return value;
            }
        };
        const compare = (data: RowData, dataCmp: RowData, keys: Array<string>) => keys.reduce((has, key) => {
            return (has && (data[key] || '') == dataCmp[key]);
        }, true);

        const { Itens } = this.getState().widget;
        const gridNames = [...Itens.map(({ ClasseCadastro }) => ClasseCadastro)];

        return broadcastManager
            .subscription('save')
            .pipe(filter(({ data }) => gridNames.includes(data?.Class || '')))
            .subscribe((message) => {
                const { Class, data, chaveprimaria } = message.data as SaveDataType;
                const Item = Itens.find(({ ClasseCadastro, DataSet }) => ClasseCadastro == Class || DataSet === 'nClientDataSetMem')!;
                const { DataSet } = Item;
                const gridInstance = gridManager.get(DataSet)!;

                const columns: Array<Column> = gridInstance
                    .option('columns')
                    .filter(({ dataField }: Column) => dataField && dataField != 'INDICE');
                    
                const newData = Object.fromEntries(
                    columns.map((column) => {
                        const { dataField } = column;

                        return [dataField, formatTypeForGrid(column, data[dataField!] || '')];
                    })
                );

                const item = this._gridData.get(DataSet)?.find((rowData) => compare(rowData, newData, chaveprimaria));

                if (item) {
                    const gridDataSource = gridInstance.getDataSource();
                    const gridStore = gridDataSource.store() as array_store;

                    this._updateDataSource(gridDataSource, gridStore.keyOf(item), newData);
                }
                else {
                    const gridDataSource = gridInstance.getDataSource();
                    const gridStore = gridDataSource.store() as array_store;

                    if (gridStore.key().includes('INDICE')) {
                        gridStore
                            .totalCount({})
                            .then((totalCount) => {
                                this._insertInDataSource(gridDataSource, {
                                    ...newData,
                                    INDICE: `${totalCount + 1}`,
                                });
                            });
                    }
                    else {
                        this._insertInDataSource(gridDataSource, newData);
                    }
                }
            });
    }

    protected _build(): JQuery<HTMLElement> {
        const { Itens, Name, colunas } = this.getState().widget;
        const wrapper = this._element.length ? this._element : $(`<div id="widget_${Name}" class="${colunas ? `col-md-${colunas}` : ''}"><div></div></div>`).css({ 'display': 'grid' });
        const treeviewEl = this._element.length ? this._element.find('div').get(0) : wrapper.find('div').get(0);
        const datasetNames = Itens.map(({ DataSet }) => DataSet);

        this._subscriptions?.forEach((subscription) => subscription.unsubscribe());
        this._subscriptions = [];

        const subscription = gridManager.afterSetGrid
            .subscribe((gridItem) => {
                if (datasetNames.includes(gridItem?.name || '')) {
                    if (datasetNames[0] !== gridItem?.name) {
                        gridItem?.instance.option('visible', false);

                        $(gridItem?.instance.element() as any).closest('[id*=widget_]')?.hide();
                    }

                    gridItem?.instance.on('contentReady', ({ component }: any) => {
                        this._gridData.set(gridItem?.name || '', GridActionHelper.getAllData(component).map((e: any) => ({ ...e })));

                        this._createTreeView(treeviewEl!);
                    });
                }
            });

        this._subscriptions.push(subscription);

        return wrapper;
    }

    private _updateDataSource(dataSource: data_source, key: number | string | Object, data: Object) {
        return new Promise((resolve, reject) => {
            const store = dataSource.store() as array_store;

            store
                .update(key, data)
                .then(() => {
                    dataSource
                        .reload()
                        .then(() => resolve(null))
                        .catch((e) => reject(e));
                })
                .catch((e) => reject(e));
        });
    }

    private _insertInDataSource(dataSource: data_source, data: object) {
        return new Promise((resolve, reject) => {
            const store = dataSource.store() as array_store;

            store
                .insert(data)
                .then(() => {
                    dataSource
                        .reload()
                        .then(() => resolve(null))
                        .catch((e) => reject(e));
                })
                .catch((e) => reject(e));
        })
    }

    private _createTreeView(element: HTMLElement) {
        const dataSource = this._createTreeViewDataSource();

        if (this._treeview) {
            this._treeview.beginUpdate();
            this._treeview.option('dataSource', dataSource);

            this._expandedNodes.forEach(item => {
                const store = this._treeview?.getDataSource().store();

                if (store?._array?.find((arrayItem: any) => arrayItem.treeviewIndex == item.treeviewIndex)) {
                    this._treeview?.expandItem(item.treeviewIndex);
                }
            });
            
            this._treeview.repaint();
            this._treeview.endUpdate();

            return;
        }

        const treeview = new dxTreeView(element, {
            dataSource: dataSource,
            selectByClick: true,
            selectionMode: 'single',
            dataStructure: 'plain',
            expandEvent: 'click',
            animationEnabled: false,
            keyExpr: 'treeviewIndex',
            focusStateEnabled: false,

            onItemClick: ({ itemData }) => {
                const { Itens } = this.getState().widget;
                const grids = Itens.map(({ DataSet }) => gridManager.get(DataSet));
                const gridToShowIndex = itemData.dataSetIndex + (itemData.isHome ? 0 : 1);

                this._expandedNodes = [{...itemData}];

                grids.forEach((grid, index) => {
                    const gridWrapper = $(grid?.element() as any).closest('[id*=widget_]');
                    if (index == gridToShowIndex) {
                        gridWrapper.show();

                        grid?.option('visible', true);
                        grid?.repaint();

                        if (itemData.isHome) {
                            grid?.clearFilter();
                        }
                        else {
                            const { DetailFields, MasterFields } = Itens[index];

                            const filter = MasterFields
                                .split(';')
                                .reduce((filter, fieldName, index) => {
                                    const detailField = DetailFields.split(';')[index];

                                    if (filter.length) {
                                        filter.push('and');
                                    }

                                    filter.push([detailField, '=', itemData[fieldName]]);

                                    return filter;
                                }, [] as Array<string | Array<string>>)

                            grid?.filter(filter);
                        }
                    }
                    else {
                        gridWrapper.hide();
                        grid?.option('visible', false);

                    }
                })

            },
            onItemExpanded: ({ component, itemData }) => {// allow only one opened branch
                const nodes = component?.getNodes();

                component?.beginUpdate();

                nodes?.forEach((node) => {
                    if (node.expanded && node.itemData.treeviewIndex != itemData.treeviewIndex) {
                        console.log('collapse', node.itemData.treeviewIndex, node.itemData, itemData);
                        component?.collapseItem(node.itemData);
                    }
                });

                component?.endUpdate();

                this._expandedNodes = [{...itemData}];
            }
        });

        this._treeview = treeview;

        return this._treeview.element();
    }

    private _getText(rowData: RowData, keys: string) {
        return keys
            .split(';')
            .map((fieldName) => (rowData[fieldName] || 'Descrição não encontrada').replace(/&nbsp/g, ' '))
            .join(' ');
    }

    private _createTreeViewDataSource() {
        const { Itens, Titulo } = this.getState().widget;
        const dependencies: ItemDependecyMap =  this._getItemDependecies();
        const arrayForDataSource: Array<RowData> = [{
            treeviewIndex: 0,
            dataSetIndex: 0,
            text: Titulo,
            isHome: true,
            icon: 'las la-home',
        }];

        const getParentId = (item: RowData, currentItem: Itens, parentItem: Itens): number => {
            const itemKey = this._getRowKey(item, currentItem.DetailFields);

            const parentData = dependencies
                .get(parentItem.DataSet)!
                .get(itemKey);

            if (Array.isArray(parentData)) {
                const parentRowData = parentData.find((data) => {
                    const parentKey = this._getRowKey(data, parentItem.MasterFields);

                    return parentKey == itemKey;
                });

                if (parentRowData == undefined) throw new Error(`Item does not have parent`);

                return parentRowData.treeviewIndex;
            }
            else {
                if (!parentData) {
                    console.log(`Item does not have parent`);

                    return -1;
                }

                return parentData.treeviewIndex;
            }
        }

        Itens.forEach(({ DataSet }, index) => {
            const dataSetMap = dependencies.get(DataSet)!;

            if (!dataSetMap) return;
            if (!index) {
                dataSetMap.forEach((rowItem: RowData) => arrayForDataSource.push({ ...rowItem }));
            }
            else {
                dataSetMap.forEach((rowData) => {
                    if (Array.isArray(rowData)) {
                        rowData.forEach((rowItem) => {
                            arrayForDataSource.push({
                                ...rowItem,
                                parentId: getParentId(rowItem, Itens[index], Itens[index - 1])
                            });
                        });
                    }
                    else {
                        arrayForDataSource.push(rowData);
                    }
                });
            }
        });

        this._dataSource = new data_source({
            store: {
                type: "array",
                key: "treeviewIndex",
                data: arrayForDataSource,
            } as DevExpress.data.ArrayStoreOptions,
        });

        return this._dataSource;
    }

    private _getItemDependecies(): ItemDependecyMap {
        const { Itens } = this.getState().widget;
        const dependencyMap: ItemDependecyMap = new Map<string, Map<string, RowData | RowData[]>>();

        for (let index = 0; index < Itens.length - 1; index++) { // dont show last level
            const { DataSet, DetailFields, ExpressaoTitulo } = Itens[index];
            const parentConfig = Itens[index - 1];
            const parentData = parentConfig ? this._gridData.get(parentConfig.DataSet) : [];
            const parentLastIndex = (parentData || []).length;
            const currentDatasetData = this._gridData.get(DataSet) || [];
            const siblings = new Map<string, RowData | Array<RowData>>();

            dependencyMap.set(DataSet, siblings);

            currentDatasetData?.forEach((rowData, dataIndex) => { // in-place changes
                const key = this._getRowKey(rowData, DetailFields);
                const siblingsData = siblings.get(key);

                rowData.treeviewIndex = dataIndex + 1 + parentLastIndex;
                rowData.dataSetName = DataSet;
                rowData.dataSetIndex = index;
                rowData.text = this._getText(rowData, ExpressaoTitulo);
                
                if (siblingsData) {
                    if (Array.isArray(siblingsData)) siblingsData.push(rowData)
                    else siblings.set(key, [siblingsData, rowData]);
                }
                else {
                    siblings.set(key, rowData);
                }
            });
        }

        return dependencyMap;
    }

    private _getRowKey(rowData: RowData, keys: string = '') {

        if (keys) {

            return keys.split(';')
            .map((fieldName) => rowData[fieldName] || '')
            .join(';');
        }

        return Object.keys(rowData).map((fieldName) => rowData[fieldName] || '')
        .join(';');
    }

    public unsubscribeBroadCast() {
        this._broadCastSubscription.unsubscribe();
    }
}