import gridManager from "../grid-manager/grid-manager";
import { Subject } from "rxjs";
import { filter } from "rxjs/operators";
import { FormatErrorMessage } from "../../../js/nDialog";
import { bufferDebounce } from "./util/buffer-debounce";
import { GridActionHelper } from "../../../js/nGrid";
import { verifyAxiosError, verifyAxiosRedirect } from "../../../js/nRequests";
import { AxiosError } from "axios";

declare var axios: any;

interface ExpressionItem {
    datasetName: string;
    widgetName: string;
    expression: string;
    className: string;
}

interface ExpressionItemWithType extends ExpressionItem {
    type: ExpressionType;
}

type ExpressionType = 'value'; // preparado para aceitar futuramente value | visible | cor ...


interface RequestExpressionValue {
    ClassName: string;
    Dataset: Array<{
        NomeDataset: string;
        Registros: Array<{
            [key: string]: string;
        }>;
    }>;
    Totalizadores: Array<{
        DataSet: string;
        WidgetName: string;
        ExpressaoValue: string;
        Valor?: string;
        Erro?: string;
    }>;
}

interface ResponseExpressionValue extends RequestExpressionValue { }


export class WidgetExpressionExecutor {
    private _requests = new Subject<ExpressionItemWithType>();
    private _externalReponseEvent = new Subject<ExpressionItem & { value: string }>();

    constructor() {
        this._setExternalRequestEvent();
    }

    private _setExternalRequestEvent() {
        this._requests
            .pipe(bufferDebounce(200))
            .subscribe((requests) => {
                this._runBlock(requests);
            })
    }

    private _runBlock(items: ExpressionItemWithType[]) {
        const groupedItems = this._groupBlockExpression(items, 'type');

        Object
            .entries(groupedItems)
            .forEach(([type, items]) => {
                switch (type as ExpressionType) {
                    case 'value':
                        this.runExternalExpressionValue(items);
                }
            });
    }

    private runExternalExpressionValue(items: ExpressionItemWithType[]) {
        const groupedByClassName = this._groupBlockExpression(items, 'className');

        Object
            .entries(groupedByClassName)
            .forEach(async ([_, expressionItems]) => {
                const groupedByDataset = this._groupBlockExpression(expressionItems, 'datasetName');

                try {
                    const { data: { data } } = await axios.post('nHttpServicos/ExecutarTotalizacoesDataset', {
                        p: {
                            ClassName: items[0].className,
                            Totalizadores: expressionItems.map(({ datasetName, widgetName, expression }) => ({
                                DataSet: datasetName,
                                ExpressaoValue: expression,
                                WidgetName: widgetName,
                            })),
                            Dataset: Object
                                .keys(groupedByDataset)
                                .map((datasetName) => ({
                                    NomeDataset: datasetName,
                                    Registros: this._getDataSetData(datasetName),
                                })),
                        }
                    } as { p: RequestExpressionValue });


                    (data as ResponseExpressionValue).Totalizadores.forEach((widgetRes) => {
                        if (widgetRes.Erro) throw new Error(widgetRes.Erro);

                        this._externalReponseEvent.next({
                            className: (data as ResponseExpressionValue).ClassName,
                            datasetName: widgetRes.DataSet,
                            expression: widgetRes.ExpressaoValue,
                            widgetName: widgetRes.WidgetName,
                            value: typeof widgetRes.Valor !== 'undefined' ? String(widgetRes.Valor) : '',
                        });
                    });
                }
                catch (e) {
                    if (verifyAxiosError(e as AxiosError)) return;

                    FormatErrorMessage('Não foi possível carregar os dados de totalizadores');

                    console.error(e);
                }
            });
    }

    private _groupBlockExpression(items: ExpressionItemWithType[], groupdBy: keyof ExpressionItemWithType) {
        return items.reduce((group, item) => {
            (group[item[groupdBy]] = group[item[groupdBy]] || []).push(item);

            return group;
        }, {} as {
            [key: string]: ExpressionItemWithType[];
        });
    }

    private _getDataSetData(gridName: string) {
        const grid = gridManager.get(gridName)!;

        try {
            return (GridActionHelper.getAllWithEditedData(grid) as any[]);
        }
        catch(e) {
            console.error(e);

            return [];
        }
    }

    public runExternal(item: ExpressionItemWithType) {
        this._requests.next(item);
    }

    public listenForExternalResponse(widgetName: string) {
        return this._externalReponseEvent.pipe(filter((a) => (a.widgetName == widgetName)));
    }
}