import dxDataGrid, { Column } from 'devextreme/ui/data_grid';
import { formatWithDataType } from '../../../js/nGrid';
import { Styles, ColumnInput, RowInput } from 'jspdf-autotable'
import { FormatErrorMessage } from '../../../js/nDialog';

interface RowData {
    [key: string]: string
}

interface columnStyles {
    [key: string]: Partial<Styles>;
}

interface ExportPdfProps {
    instance: dxDataGrid;
    data: Array<RowData>;
    fileName?: string;
}

export const exportPdf = ({ data, instance, fileName = 'pdf' }: ExportPdfProps) => {
    return Promise
        .all([
            import('jspdf'),
            import('jspdf-autotable')
        ])
        .then(([{ default: jsPDF }, { default: autoTable }]) => {
            const gridColumns = instance.getVisibleColumns().filter(column => !column.type || column.type == 'groupExpand');

            const doc = new jsPDF({
                orientation: 'landscape',
                compress: true,
            });

            autoTable(doc, {
                columns: getColumns(gridColumns),
                body: getRows(gridColumns, data),
                columnStyles: getColumnStyles(gridColumns),
                headStyles:{
                    fontStyle: 'bold',
                    fontSize: 4,
                    textColor: 'white',
                    fillColor: '#2d4154',
                    halign: 'center',
                },
                styles: {
                    fontSize: 4,
                    cellWidth: 'auto',
                    textColor: 'black',
                },
            });

            doc.save(`${fileName}`);
        })
        .catch((e) => {
            console.error(e);

            FormatErrorMessage('Falha ao gerar pdf. Tente novamente!');
        });
};

const getColumns = (gridColumns: Array<Column>): Array<ColumnInput> => gridColumns.map(column => ({
    header: column.caption || '',
    dataKey: column.dataField,
}));

const getRows = (gridColumns: Array<Column>, data: Array<RowData>): Array<RowInput> => data.map((row) => formatRowForPdf(gridColumns, row));

const formatRowForPdf = (gridColumns: Array<Column>, row: RowData) => {
    const visibleRowValues = gridColumns.reduce((rowValues, column) => ({
        ...rowValues,
        [column.dataField!]: row[column.dataField!],
    }), {} as RowData);;

    Object
        .entries(visibleRowValues)
        .forEach(([columnName, columnValue]) => {
            const { dataType } = gridColumns.find(column => column.dataField == columnName) || {};

            if (!dataType) return;

            formatWithDataType(dataType!, visibleRowValues, columnName, columnValue);

            if (typeof columnValue === "boolean" || (columnValue as any) instanceof Boolean) {
                columnValue = (columnValue) ? "Ativo" : "Inativo";
            }

            // strip html from columns
            visibleRowValues[columnName] = visibleRowValues[columnName].replace(/<(?:.|\n)*?>/gm, "").trim();
        });

    return visibleRowValues;
};

const getColumnStyles = (gridColumns: Array<Column>) => {
    return gridColumns.reduce((styles, column) => ({
        ...styles,
        [column.dataField as string]: {
            halign: column.alignment,
        },
    }), {} as columnStyles);
};