import { CleanWebpackPlugin } from 'clean-webpack-plugin';
import dotenv from 'dotenv';
import MiniCssExtractPlugin from 'mini-css-extract-plugin';
import path from 'path';
import { Configuration } from "webpack";
import { merge } from 'webpack-merge';
import devConfig from './webpack.dev';
import prodConfig from './webpack.prod';
import { GenerateMustacheScriptsPartial } from './WebpackUtils/GenerateMustacheScriptsPartial';

const ConcatPlugin = require('@mcler/webpack-concat-plugin');

const ensureEnvMode = (mode?: string): "development" | "production" => mode == 'development' ? mode : 'production';

dotenv.config({ path: '.env' });
dotenv.config({ path: 'development.env' });
dotenv.config({ path: 'production.env' });

const inProduction = ensureEnvMode(process.env.NODE_ENV) === 'production';

console.log("Mode: ", ensureEnvMode(process.env.NODE_ENV));

const legacyLoaderFiles = [
    new ConcatPlugin({
        name: 'external-vendors-head',
        fileName: '[name].[hash].js',
        filesToConcat: [
            './metronic/demo1/plugins/global/plugins.bundle.js',
            './vendors/moment/lang/pt-br.js',
            './vendors/jQuery-ContextMenu/jquery.contextMenu.min.js',
            './vendors/jQuery-ContextMenu/jquery.ui.position.js',
        ],
        attributes: {
            async: true
        }
    }),
    new ConcatPlugin({
        name: 'external-vendors-head',
        fileName: '[name].[hash].css',
        filesToConcat: [
            './metronic/demo1/plugins/custom/fullcalendar/fullcalendar.bundle.css',
            './metronic/demo1/css/style.bundle.css',
			// begin::metronicLayout Skins(used by all pages) 
            './metronic/demo1/css/themes/layout/header/base/dark.css',
            './metronic/demo1/css/themes/layout/brand/dark.css',
            './metronic/demo1/css/pages/wizard/wizard-3.css',
            './metronic/demo1/css/pages/wizard/wizard-2.css',
            './metronic/demo1/css/pages/wizard/wizard-1.css',
            './metronic/demo1/css/pages/login/classic/login-1.css',
			// end::metronicLayout Skins(used by all pages) 
            './css/dashboard/dashboard.min.css',
            './vendors/introjs/introjs.min.css',
            './vendors/jQuery-ContextMenu/jquery.contextMenu.min.css',
        ],
        attributes: {
            async: true
        }
    }),
    new ConcatPlugin({
        name: 'external-vendors-body',
        fileName: '[name].[hash].js',
        filesToConcat: [
            './metronic/demo1/js/scripts.bundle.js',
            './vendors/introjs/intro.min.js',
            './vendors/jQueryMaskPlugin/jquery.mask.js',
        ],
        attributes: {
            async: true
        }
    }),
];

const config: Configuration = {
    mode: ensureEnvMode(process.env.NODE_ENV),
    entry: {
        qcloud: ['./js/modulesImporter.js']
    },
    output: {
        path: path.resolve(__dirname, 'dist'),
        publicPath: '.static/dist/',
        filename: '[name].[contenthash].bundle.js',
        sourceMapFilename: 'sourceMap/[name].[contenthash].js.map',
        chunkFilename: 'chunk/[name].[contenthash].js'
    },
    context: path.resolve(__dirname, '../.static'),
    optimization: {
        minimize: true,
        splitChunks: {
            chunks: "all",
            cacheGroups: {
                devextreme: {
                    test: /[\\/]node_modules[\\/]devextreme[\\/]/,
                    name: 'npm.devextreme'
                },
                pdf: {
                    test: /[\\/]node_modules[\\/](jspdf|html2canvas|jspdf-autotable)[\\/]/,
                    name: 'npm.jspdf'
                },
                jstree: {
                    test: /[\\/]node_modules[\\/](jstree)[\\/]/,
                    name: 'npm.jstree'
                },
                vendors: {
                    test: /[\\/]node_modules[\\/](?!.*devextreme|jspdf|html2canvas|jstree).*/, // ignore devextreme folder
                    name: 'npm.vendors'
                },
            }
        }
    },
    resolve: {
        extensions: ['.ts', '.js']
    },
    performance: {
        maxEntrypointSize: 3145728,
        maxAssetSize: 3145728
    },
    plugins: [
        new CleanWebpackPlugin(),
        new MiniCssExtractPlugin({
            filename: '[name].[contenthash].bundle.css',
        }),
        ...legacyLoaderFiles,
        new GenerateMustacheScriptsPartial(),
    ],
    module: {
        strictExportPresence: true,
        rules: [
            {
                test: /\.ts$/,
                loader: 'ts-loader',
                exclude: [/node_modules/]
            },
            {
                test: require.resolve('jquery'),
                use: [
                    {
                        loader: 'expose-loader',
                        options: {
                            exposes: 'jQuery'
                        }
                    },
                    {
                        loader: 'expose-loader',
                        options: {
                            exposes: '$'
                        }
                    }
                ]
            },
            {
                test: /\.js$/,
                exclude: /(node_modules)/,
                use: {
                    loader: 'babel-loader',
                    options: {
                        cacheDirectory: true
                    }
                }
            },
            {
                test: /\.html$/i,
                loader: 'html-loader'
            },
            {
                test: /\.css$/i,
                use: [MiniCssExtractPlugin.loader, 'css-loader']
            },
            {
                test: /\.(woff|woff2|eot|ttf|otf)$/i,
                type: 'asset/resource',
                generator: {
                    outputPath: '.static/dist/',
                    filename: 'fonts/[hash][ext][query]'
                }
            },
            {
                test: /\.(png|svg|jpg|jpeg|gif)$/i,
                type: 'asset/resource',
                generator: {
                    outputPath: '.static/dist/',
                    filename: 'images/[hash][ext][query]'
                }
              },
            {
                test: /\.scss$/i,
                use: [
                    'style-loader',
                    {
                        loader: 'css-loader',
                        options: {
                            importLoaders: 1,
                            modules: {
                                localIdentName: '[local]__[hash:base64]'
                            },
                        }
                    },
                    {
                        loader: 'sass-loader',
                        options: {
                            sourceMap: true,
                        }

                    }
                ]
            }
        ]
    },
    externals: [
        {
            jquery: 'jQuery',
            moment: 'moment', // included by metronic
            toastr: 'toastr', // included by metronic
            sweetalert2: 'Swal' // included by metronic
        },
    ]
};

export default merge(inProduction ? prodConfig : devConfig, config);